from __future__ import annotations

import itertools
import logging
import os
import unittest
from datetime import datetime, timedelta, timezone
from threading import Event
from typing import NamedTuple
from unittest.mock import Mock, call, patch

import freezegun
import pytest
import requests_mock as rm
from requests.exceptions import InvalidSchema

from streamlink.session import Streamlink
from streamlink.stream.hls import (
    M3U8,
    HLSPlaylist,
    HLSSegment,
    HLSStream,
    HLSStreamReader,
    M3U8Parser,
    MuxedHLSStream,
)
from streamlink.stream.hls.hls import log
from streamlink.utils.crypto import AES, pad
from tests.mixins.stream_hls import EventedHLSStreamWorker, EventedHLSStreamWriter, Playlist, Segment, Tag, TestMixinStreamHLS
from tests.resources import text


EPOCH = datetime(2000, 1, 1, 0, 0, 0, 0, tzinfo=timezone.utc)
ONE_SECOND = timedelta(seconds=1.0)


class EncryptedBase:
    content: bytes
    content_plain: bytes

    def __init__(self, num, key, iv, *args, content=None, padding=b"", append=b"", **kwargs):
        super().__init__(num, *args, **kwargs)
        aesCipher = AES.new(key, AES.MODE_CBC, iv)
        content = self.content if content is None else content
        padded = content + padding if padding else pad(content, AES.block_size, style="pkcs7")
        self.content_plain = content
        self.content = aesCipher.encrypt(padded) + append


class TagMap(Tag):
    def __init__(self, num, namespace, attrs=None):
        self.path = f"map{num}"
        self.content = f"[map{num}]".encode("ascii")
        super().__init__(
            "EXT-X-MAP",
            {
                "URI": self.val_quoted_string(self.url(namespace)),
                **(attrs or {}),
            },
        )


class TagMapEnc(EncryptedBase, TagMap):
    pass


class TagKey(Tag):
    _id = itertools.count()

    def __init__(self, method="NONE", uri=None, iv=None, keyformat=None, keyformatversions=None):
        attrs = {"METHOD": method}
        if uri is not False:  # pragma: no branch
            attrs["URI"] = lambda tag, namespace: tag.val_quoted_string(tag.url(namespace))
        if iv is not None:  # pragma: no branch
            attrs["IV"] = self.val_hex(iv)
        if keyformat is not None:  # pragma: no branch
            attrs["KEYFORMAT"] = self.val_quoted_string(keyformat)
        if keyformatversions is not None:  # pragma: no branch
            attrs["KEYFORMATVERSIONS"] = self.val_quoted_string(keyformatversions)
        super().__init__("EXT-X-KEY", attrs)
        self.uri = uri
        self.path = f"encryption{next(self._id)}.key"

    def url(self, namespace):
        return self.uri.format(namespace=namespace) if self.uri else super().url(namespace)


class SegmentEnc(EncryptedBase, Segment):
    pass


def test_logger_name():
    assert log.name == "streamlink.stream.hls"


def test_repr(session: Streamlink):
    stream = HLSStream(session, "https://foo.bar/playlist.m3u8")
    assert repr(stream) == "<HLSStream ['hls', 'https://foo.bar/playlist.m3u8']>"

    stream = HLSStream(session, "https://foo.bar/playlist.m3u8", "https://foo.bar/master.m3u8")
    assert repr(stream) == "<HLSStream ['hls', 'https://foo.bar/playlist.m3u8', 'https://foo.bar/master.m3u8']>"


class TestHLSVariantPlaylist:
    @pytest.fixture()
    def streams(self, request: pytest.FixtureRequest, requests_mock: rm.Mocker, session: Streamlink):
        url = f"http://mocked/{request.node.originalname}/master.m3u8"
        playlist = getattr(request, "param", "")

        with text(playlist) as fd:
            content = fd.read()
        requests_mock.get(url, text=content)

        return HLSStream.parse_variant_playlist(session, url)

    @pytest.mark.parametrize("streams", ["hls/test_master.m3u8"], indirect=True)
    def test_variant_playlist(self, request: pytest.FixtureRequest, streams: dict[str, HLSStream]):
        assert list(streams.keys()) == ["720p", "720p_alt", "480p", "360p", "160p", "1080p (source)", "90k"]
        assert all(isinstance(stream, HLSStream) for stream in streams.values())
        assert all(stream.multivariant is not None and stream.multivariant.is_master for stream in streams.values())

        base = f"http://mocked/{request.node.originalname}"
        stream = next(iter(streams.values()))
        assert repr(stream) == f"<HLSStream ['hls', '{base}/720p.m3u8', '{base}/master.m3u8']>"

        assert stream.multivariant is not None
        assert stream.multivariant.uri == f"{base}/master.m3u8"
        assert stream.url_master == f"{base}/master.m3u8"

    def test_url_master(self, session: Streamlink):
        stream = HLSStream(session, "http://mocked/foo", url_master="http://mocked/master.m3u8")

        assert stream.multivariant is None
        assert stream.url == "http://mocked/foo"
        assert stream.url_master == "http://mocked/master.m3u8"


class EventedWorkerHLSStreamReader(HLSStreamReader):
    __worker__ = EventedHLSStreamWorker


class EventedWriterHLSStreamReader(HLSStreamReader):
    __writer__ = EventedHLSStreamWriter


class EventedWorkerHLSStream(HLSStream):
    __reader__ = EventedWorkerHLSStreamReader


class EventedWriterHLSStream(HLSStream):
    __reader__ = EventedWriterHLSStreamReader


@patch("streamlink.stream.hls.hls.HLSStreamWorker.wait", Mock(return_value=True))
class TestHLSStream(TestMixinStreamHLS, unittest.TestCase):
    def get_session(self, options=None, *args, **kwargs):
        session = super().get_session(options)
        session.set_option("hls-live-edge", 3)

        return session

    def test_thread_names(self):
        self.subject(playlists=[Playlist(0, [Segment(0)], end=True)])
        assert self.thread.reader.worker.name == "HLSStreamWorker-0"
        assert self.thread.reader.writer.name == "HLSStreamWriter-0"
        assert self.thread.reader.writer.executor._thread_name_prefix == "HLSStreamWriter-0-executor"
        self.await_read(read_all=True)

    def test_playlist_end(self):
        segments = self.subject([
            Playlist(0, [Segment(0)], end=True),
        ])

        assert self.await_read(read_all=True) == self.content(segments), "Stream ends and read-all handshake doesn't time out"

    def test_playlist_end_on_empty_reload(self):
        segments = self.subject([
            Playlist(0, [Segment(0)]),
            Playlist(0, [Segment(0)], end=True),
        ])

        assert self.await_read(read_all=True) == self.content(segments), "Stream ends and read-all handshake doesn't time out"

    def test_offset_and_duration(self):
        segments = self.subject(
            [
                Playlist(1234, [Segment(0), Segment(1, duration=0.5), Segment(2, duration=0.5), Segment(3)], end=True),
            ],
            streamoptions={"start_offset": 1, "duration": 1},
        )

        data = self.await_read(read_all=True)
        assert data == self.content(segments, cond=lambda s: 0 < s.num < 3), "Respects the offset and duration"
        assert all(self.called(s) for s in segments.values() if 0 < s.num < 3), "Downloads second and third segment"
        assert not any(self.called(s) for s in segments.values() if 0 > s.num > 3), "Skips other segments"

    def test_map(self):
        discontinuity = Tag("EXT-X-DISCONTINUITY")
        map1 = TagMap(1, self.id())
        map2 = TagMap(2, self.id())
        self.mock("GET", self.url(map1), content=map1.content)
        self.mock("GET", self.url(map2), content=map2.content)

        segments = self.subject(
            [
                Playlist(0, [map1, Segment(0), Segment(1), Segment(2), Segment(3)]),
                Playlist(4, [map1, Segment(4), map2, Segment(5), Segment(6), discontinuity, map1, Segment(7)], end=True),
            ],
            options={"stream-segment-threads": 2},
        )

        data = self.await_read(read_all=True, timeout=None)
        assert data == self.content([
            map1, segments[1], segments[2], segments[3],
            segments[4], map2, segments[5], segments[6], map1, segments[7],
        ])  # fmt: skip
        assert self.called(map1, once=True), "Downloads first map only once"
        assert self.called(map2, once=True), "Downloads second map only once"


# TODO: finally rewrite the segmented/HLS test setup using pytest and replace redundant setups with parametrization
@patch("streamlink.stream.hls.hls.HLSStreamWorker.wait", Mock(return_value=True))
class TestHLSStreamPlaylistReloadDiscontinuity(TestMixinStreamHLS, unittest.TestCase):
    @patch("streamlink.stream.hls.hls.log")
    def test_no_discontinuity(self, mock_log: Mock):
        segments = self.subject([
            Playlist(0, [Segment(0), Segment(1)]),
            Playlist(2, [Segment(2), Segment(3)]),
            Playlist(4, [Segment(4), Segment(5)], end=True),
        ])

        data = self.await_read(read_all=True)
        assert data == self.content(segments)
        assert all(self.called(s) for s in segments.values())
        assert mock_log.warning.call_args_list == []

    @patch("streamlink.stream.hls.hls.log")
    def test_discontinuity_single_segment(self, mock_log: Mock):
        segments = self.subject([
            Playlist(0, [Segment(0), Segment(1)]),
            Playlist(2, [Segment(2), Segment(3)]),
            Playlist(5, [Segment(5), Segment(6)]),
            Playlist(8, [Segment(8), Segment(9)], end=True),
        ])

        data = self.await_read(read_all=True)
        assert data == self.content(segments)
        assert all(self.called(s) for s in segments.values())
        assert mock_log.warning.call_args_list == [
            call("Skipped segment 4 after playlist reload. This is unsupported and will result in incoherent output data."),
            call("Skipped segment 7 after playlist reload. This is unsupported and will result in incoherent output data."),
        ]

    @patch("streamlink.stream.hls.hls.log")
    def test_discontinuity_multiple_segments(self, mock_log: Mock):
        segments = self.subject([
            Playlist(0, [Segment(0), Segment(1)]),
            Playlist(2, [Segment(2), Segment(3)]),
            Playlist(6, [Segment(6), Segment(7)]),
            Playlist(10, [Segment(10), Segment(11)], end=True),
        ])

        data = self.await_read(read_all=True)
        assert data == self.content(segments)
        assert all(self.called(s) for s in segments.values())
        assert mock_log.warning.call_args_list == [
            call("Skipped segments 4-5 after playlist reload. This is unsupported and will result in incoherent output data."),
            call("Skipped segments 8-9 after playlist reload. This is unsupported and will result in incoherent output data."),
        ]


class TestHLSStreamWorker(TestMixinStreamHLS, unittest.TestCase):
    __stream__ = EventedWorkerHLSStream

    OPTIONS = {"stream-timeout": 1}

    def tearDown(self) -> None:
        worker: EventedHLSStreamWorker = self.thread.reader.worker  # type: ignore[assignment]
        # don't await the handshakes on error
        worker.handshake_wait.go()
        worker.handshake_reload.go()
        return super().tearDown()

    def get_session(self, options=None, *args, **kwargs):
        return super().get_session({**self.OPTIONS, **(options or {})}, *args, **kwargs)

    def test_segment_queue_timing_threshold_reached(self) -> None:
        self.subject(
            start=False,
            playlists=[
                Playlist(0, targetduration=5, segments=[Segment(0)]),
                # no EXT-X-ENDLIST, last mocked playlist response will be repreated forever
                Playlist(0, targetduration=5, segments=[Segment(0), Segment(1)]),
            ],
        )
        worker: EventedHLSStreamWorker = self.thread.reader.worker  # type: ignore[assignment]
        targetduration = ONE_SECOND * 5

        with (
            freezegun.freeze_time(EPOCH) as frozen_time,
            patch("streamlink.stream.hls.hls.log") as mock_log,
        ):
            self.start()

            assert worker.handshake_reload.wait_ready(1), "Loads playlist for the first time"
            assert worker.playlist_sequence == -1, "Initial sequence number"
            assert worker.playlist_sequence_last == EPOCH, "Sets the initial last queue time"

            # first playlist reload has taken one second
            frozen_time.tick(ONE_SECOND)
            self.await_playlist_reload(1)

            assert worker.handshake_wait.wait_ready(1), "Arrives at wait() call #1"
            assert worker.playlist_sequence == 1, "Updates the sequence number"
            assert worker.playlist_sequence_last == EPOCH + ONE_SECOND, "Updates the last queue time"
            assert worker.playlist_targetduration == 5.0

            # trigger next reload when the target duration has passed
            frozen_time.tick(targetduration)
            self.await_playlist_wait(1)
            self.await_playlist_reload(1)

            assert worker.handshake_wait.wait_ready(1), "Arrives at wait() call #2"
            assert worker.playlist_sequence == 2, "Updates the sequence number again"
            assert worker.playlist_sequence_last == EPOCH + ONE_SECOND + targetduration, "Updates the last queue time again"
            assert worker.playlist_targetduration == 5.0

            for num in range(3, 6):
                # trigger next reload when the target duration has passed
                frozen_time.tick(targetduration)
                self.await_playlist_wait(1)
                self.await_playlist_reload(1)

                assert worker.handshake_wait.wait_ready(1), f"Arrives at wait() call #{num}"
                assert worker.playlist_sequence == 2, "Sequence number is unchanged"
                assert worker.playlist_sequence_last == EPOCH + ONE_SECOND + targetduration, "Last queue time is unchanged"
                assert worker.playlist_targetduration == 5.0

            assert mock_log.warning.call_args_list == []

            # trigger next reload when the target duration has passed
            frozen_time.tick(targetduration)
            self.await_playlist_wait(1)
            self.await_playlist_reload(1)

            self.await_read(read_all=True)
            self.await_close(1)

            assert mock_log.warning.call_args_list == [call("No new segments in playlist for more than 15.00s. Stopping...")]

    def test_segment_queue_timing_threshold_reached_ignored(self) -> None:
        segments = self.subject(
            start=False,
            options={"hls-segment-queue-threshold": 0},
            playlists=[
                # no EXT-X-ENDLIST, last mocked playlist response will be repreated forever
                Playlist(0, targetduration=5, segments=[Segment(0)]),
            ],
        )
        worker: EventedHLSStreamWorker = self.thread.reader.worker  # type: ignore[assignment]
        targetduration = ONE_SECOND * 5

        with freezegun.freeze_time(EPOCH) as frozen_time:
            self.start()

            assert worker.handshake_reload.wait_ready(1), "Loads playlist for the first time"
            assert worker.playlist_sequence == -1, "Initial sequence number"
            assert worker.playlist_sequence_last == EPOCH, "Sets the initial last queue time"

            # first playlist reload has taken one second
            frozen_time.tick(ONE_SECOND)
            self.await_playlist_reload(1)

            assert worker.handshake_wait.wait_ready(1), "Arrives at first wait() call"
            assert worker.playlist_sequence == 1, "Updates the sequence number"
            assert worker.playlist_sequence_last == EPOCH + ONE_SECOND, "Updates the last queue time"
            assert worker.playlist_targetduration == 5.0
            assert self.await_read() == self.content(segments)

            # keep reloading a couple of times
            for num in range(10):
                frozen_time.tick(targetduration)
                self.await_playlist_wait(1)
                self.await_playlist_reload(1)

                assert worker.handshake_wait.wait_ready(1), f"Arrives at wait() #{num + 1}"
                assert worker.playlist_sequence == 1, "Sequence number is unchanged"
                assert worker.playlist_sequence_last == EPOCH + ONE_SECOND, "Last queue time is unchanged"

        assert self.thread.data == [], "No new data"
        assert worker.is_alive()

        # make stream end gracefully to avoid any unnecessary thread blocking
        self.thread.reader.writer.put(None)

    def test_segment_queue_timing_threshold_reached_min(self) -> None:
        self.subject(
            start=False,
            playlists=[
                # no EXT-X-ENDLIST, last mocked playlist response will be repreated forever
                Playlist(0, targetduration=1, segments=[Segment(0)]),
            ],
        )
        worker: EventedHLSStreamWorker = self.thread.reader.worker  # type: ignore[assignment]
        targetduration = ONE_SECOND

        with (
            freezegun.freeze_time(EPOCH) as frozen_time,
            patch("streamlink.stream.hls.hls.log") as mock_log,
        ):
            self.start()

            assert worker.handshake_reload.wait_ready(1), "Loads playlist for the first time"
            assert worker.playlist_sequence == -1, "Initial sequence number"
            assert worker.playlist_sequence_last == EPOCH, "Sets the initial last queue time"

            # first playlist reload has taken one second
            frozen_time.tick(ONE_SECOND)
            self.await_playlist_reload(1)

            assert worker.handshake_wait.wait_ready(1), "Arrives at wait() call #1"
            assert worker.playlist_sequence == 1, "Updates the sequence number"
            assert worker.playlist_sequence_last == EPOCH + ONE_SECOND, "Updates the last queue time"
            assert worker.playlist_targetduration == 1.0

            for num in range(2, 7):
                # trigger next reload when the target duration has passed
                frozen_time.tick(targetduration)
                self.await_playlist_wait(1)
                self.await_playlist_reload(1)

                assert worker.handshake_wait.wait_ready(1), f"Arrives at wait() call #{num}"
                assert worker.playlist_sequence == 1, "Sequence number is unchanged"
                assert worker.playlist_sequence_last == EPOCH + ONE_SECOND, "Last queue time is unchanged"
                assert worker.playlist_targetduration == 1.0

            assert mock_log.warning.call_args_list == []

            # trigger next reload when the target duration has passed
            frozen_time.tick(targetduration)
            self.await_playlist_wait(1)
            self.await_playlist_reload(1)

            # FIXME: NO-GIL
            #   Closing only the worker thread when no new segments were queued keeps the writer thread and buffer still open,
            #   which is why the test's reader thread keeps running until the test teardown,
            #   but this somehow breaks the assertion down below, so close everything manually...
            #   These tests will have to be rewritten eventually in pytest-style, without having to mock log calls.
            self.thread.close()
            self.thread.join(1)

            assert mock_log.warning.call_args_list == [call("No new segments in playlist for more than 5.00s. Stopping...")]

    def test_playlist_reload_offset(self) -> None:
        segments = self.subject(
            start=False,
            playlists=[
                Playlist(0, targetduration=5, segments=[Segment(0)]),
                Playlist(1, targetduration=5, segments=[Segment(1)]),
                Playlist(2, targetduration=5, segments=[Segment(2)]),
                Playlist(3, targetduration=5, segments=[Segment(3)]),
                Playlist(4, targetduration=5, segments=[Segment(4)], end=True),
            ],
        )
        worker: EventedHLSStreamWorker = self.thread.reader.worker  # type: ignore[assignment]
        targetduration = ONE_SECOND * 5

        with freezegun.freeze_time(EPOCH) as frozen_time:
            self.start()

            assert worker.handshake_reload.wait_ready(1), "Arrives at initial playlist reload"
            assert worker.playlist_reload_last == EPOCH, "Sets the initial value of the last reload time"

            # adjust clock and reload playlist: let it take one second
            frozen_time.move_to(worker.playlist_reload_last + ONE_SECOND)
            self.await_playlist_reload()
            assert worker.playlist_reload_time == 5.0, "Uses the playlist's targetduration as reload time"

            # time_completed = 00:00:01; time_elapsed = 1s
            assert worker.handshake_wait.wait_ready(1), "Arrives at first wait() call"
            assert worker.playlist_sequence == 1, "Has queued first segment"
            assert worker.playlist_end is None, "Stream hasn't ended yet"
            assert worker.time_wait == 4.0, "Waits for 4 seconds out of the 5 seconds reload time"
            self.await_playlist_wait()

            assert worker.handshake_reload.wait_ready(1), "Arrives at second playlist reload"
            assert worker.playlist_reload_last == EPOCH + targetduration, \
                "Last reload time is the sum of reload+wait time (=targetduration)"  # fmt: skip

            # adjust clock and reload playlist: let it exceed targetduration by two seconds
            frozen_time.move_to(worker.playlist_reload_last + targetduration + ONE_SECOND * 2)
            self.await_playlist_reload()
            assert worker.playlist_reload_time == 5.0, "Uses the playlist's targetduration as reload time"

            # time_completed = 00:00:12; time_elapsed = 7s (exceeded 5s targetduration)
            assert worker.handshake_wait.wait_ready(1), "Arrives at second wait() call"
            assert worker.playlist_sequence == 2, "Has queued second segment"
            assert worker.playlist_end is None, "Stream hasn't ended yet"
            assert worker.time_wait == 0.0, "Doesn't wait when reloading took too long"
            self.await_playlist_wait()

            assert worker.handshake_reload.wait_ready(1), "Arrives at third playlist reload"
            assert worker.playlist_reload_last == EPOCH + targetduration * 2 + ONE_SECOND * 2, \
                "Sets last reload time to current time when reloading took too long (changes the interval)"  # fmt: skip

            # adjust clock and reload playlist: let it take one second again
            frozen_time.move_to(worker.playlist_reload_last + ONE_SECOND)
            self.await_playlist_reload()
            assert worker.playlist_reload_time == 5.0, "Uses the playlist's targetduration as reload time"

            # time_completed = 00:00:13; time_elapsed = 1s
            assert worker.handshake_wait.wait_ready(1), "Arrives at third wait() call"
            assert worker.playlist_sequence == 3, "Has queued third segment"
            assert worker.playlist_end is None, "Stream hasn't ended yet"
            assert worker.time_wait == 4.0, "Waits for 4 seconds out of the 5 seconds reload time"
            self.await_playlist_wait()

            assert worker.handshake_reload.wait_ready(1), "Arrives at fourth playlist reload"
            assert worker.playlist_reload_last == EPOCH + targetduration * 3 + ONE_SECOND * 2, \
                "Last reload time is the sum of reload+wait time (=targetduration) of the changed interval"  # fmt: skip

            # adjust clock and reload playlist: simulate no fetch+processing delay
            frozen_time.move_to(worker.playlist_reload_last)
            self.await_playlist_reload()
            assert worker.playlist_reload_time == 5.0, "Uses the playlist's targetduration as reload time"

            # time_completed = 00:00:17; time_elapsed = 0s
            assert worker.handshake_wait.wait_ready(1), "Arrives at fourth wait() call"
            assert worker.playlist_sequence == 4, "Has queued fourth segment"
            assert worker.playlist_end is None, "Stream hasn't ended yet"
            assert worker.time_wait == 5.0, "Waits for the whole reload time"
            self.await_playlist_wait()

            assert worker.handshake_reload.wait_ready(1), "Arrives at fifth playlist reload"
            assert worker.playlist_reload_last == EPOCH + targetduration * 4 + ONE_SECOND * 2, \
                "Last reload time is the sum of reload+wait time (no delay)"  # fmt: skip

            # adjusting the clock is not needed anymore
            self.await_playlist_reload()
            assert self.await_read(read_all=True) == self.content(segments)
            self.await_close()
            assert worker.playlist_end == 4, "Stream has ended"
            assert not worker.handshake_wait.wait_ready(0), "Doesn't wait once ended"
            assert not worker.handshake_reload.wait_ready(0), "Doesn't reload playlist once ended"


@patch("streamlink.stream.hls.hls.HLSStreamWorker.wait", Mock(return_value=True))
class TestHLSStreamByterange(TestMixinStreamHLS, unittest.TestCase):
    __stream__ = EventedWriterHLSStream

    # The dummy segments in the error tests are required because the writer's run loop would otherwise continue forever
    # due to the segment's future result being None (no requests result), and we can't await the end of the stream
    # without waiting for the stream's timeout error. The dummy segments ensure that we can call await_write for these
    # successful segments, so we can close the stream afterward and safely make the test assertions.
    # The EventedHLSStreamWriter could also implement await_fetch, but this is unnecessarily more complex than it already is.

    @patch("streamlink.stream.hls.hls.log")
    def test_unknown_offset(self, mock_log: Mock):
        self.subject([
            Playlist(
                0,
                [
                    Tag("EXT-X-BYTERANGE", "3"),
                    Segment(0),
                    Segment(1),
                ],
                end=True,
            ),
        ])

        self.await_write(2 - 1)
        self.thread.close()

        assert mock_log.error.call_args_list == [
            call("Failed to fetch segment 0: Missing BYTERANGE offset"),
        ]
        assert not self.called(Segment(0))

    @patch("streamlink.stream.hls.hls.log")
    def test_unknown_offset_map(self, mock_log: Mock):
        map1 = TagMap(1, self.id(), {"BYTERANGE": '"1234"'})
        self.mock("GET", self.url(map1), content=map1.content)
        self.subject([
            Playlist(
                0,
                [
                    Segment(0),
                    map1,
                    Segment(1),
                ],
                end=True,
            ),
        ])

        self.await_write(3 - 1)
        self.thread.close()

        assert mock_log.error.call_args_list == [
            call("Failed to fetch map for segment 1: Missing BYTERANGE offset"),
        ]
        assert not self.called(map1)

    @patch("streamlink.stream.hls.hls.log")
    def test_invalid_offset_reference(self, mock_log: Mock):
        self.subject([
            Playlist(
                0,
                [
                    Tag("EXT-X-BYTERANGE", "3@0"),
                    Segment(0),
                    Segment(1),
                    Tag("EXT-X-BYTERANGE", "5"),
                    Segment(2),
                    Segment(3),
                ],
                end=True,
            ),
        ])

        self.await_write(4 - 1)
        self.thread.close()

        assert mock_log.error.call_args_list == [
            call("Failed to fetch segment 2: Missing BYTERANGE offset"),
        ]
        assert self.mocks[self.url(Segment(0))].last_request._request.headers["Range"] == "bytes=0-2"
        assert not self.called(Segment(2))

    def test_offsets(self):
        map1 = TagMap(1, self.id(), {"BYTERANGE": '"1234@0"'})
        map2 = TagMap(2, self.id(), {"BYTERANGE": '"42@1337"'})
        self.mock("GET", self.url(map1), content=map1.content)
        self.mock("GET", self.url(map2), content=map2.content)
        s1, s2, s3, s4, s5 = Segment(0), Segment(1), Segment(2), Segment(3), Segment(4)

        self.subject([
            Playlist(
                0,
                [
                    map1,
                    Tag("EXT-X-BYTERANGE", "5@3"),
                    s1,
                    Tag("EXT-X-BYTERANGE", "7"),
                    s2,
                    map2,
                    Tag("EXT-X-BYTERANGE", "11"),
                    s3,
                    Tag("EXT-X-BYTERANGE", "17@13"),
                    s4,
                    Tag("EXT-X-BYTERANGE", "19"),
                    s5,
                ],
                end=True,
            ),
        ])

        self.await_write(1 + 2 + 1 + 3)  # 1 map, 2 partial segments, 1 map, 3 partial segments
        self.await_read(read_all=True)
        assert self.mocks[self.url(map1)].last_request._request.headers["Range"] == "bytes=0-1233"
        assert self.mocks[self.url(map2)].last_request._request.headers["Range"] == "bytes=1337-1378"
        assert self.mocks[self.url(s1)].last_request._request.headers["Range"] == "bytes=3-7"
        assert self.mocks[self.url(s2)].last_request._request.headers["Range"] == "bytes=8-14"
        assert self.mocks[self.url(s3)].last_request._request.headers["Range"] == "bytes=15-25"
        assert self.mocks[self.url(s4)].last_request._request.headers["Range"] == "bytes=13-29"
        assert self.mocks[self.url(s5)].last_request._request.headers["Range"] == "bytes=30-48"


@patch("streamlink.stream.hls.hls.HLSStreamWorker.wait", Mock(return_value=True))
class TestHLSStreamEncrypted(TestMixinStreamHLS, unittest.TestCase):
    __stream__ = EventedWriterHLSStream

    def get_session(self, options=None, *args, **kwargs):
        session = super().get_session(options)
        session.set_option("hls-live-edge", 3)
        session.set_option("http-headers", {"X-FOO": "BAR"})

        return session

    def gen_key(
        self,
        aes_key=None,
        aes_iv=None,
        method="AES-128",
        uri=None,
        keyformat="identity",
        keyformatversions=1,
        mock=None,
    ):
        aes_key = aes_key or os.urandom(16)
        aes_iv = aes_iv or os.urandom(16)

        key = TagKey(method=method, uri=uri, iv=aes_iv, keyformat=keyformat, keyformatversions=keyformatversions)
        self.mock("GET", key.url(self.id()), **(mock if mock else {"content": aes_key}))

        return aes_key, aes_iv, key

    @patch("streamlink.stream.hls.hls.log")
    def test_hls_encrypted_invalid_method(self, mock_log: Mock):
        aesKey, aesIv, key = self.gen_key(method="INVALID")

        self.subject([
            Playlist(0, [key, SegmentEnc(1, aesKey, aesIv)], end=True),
        ])
        self.await_write()

        self.thread.close()
        self.await_close()

        assert b"".join(self.thread.data) == b""
        assert mock_log.error.mock_calls == [
            call("Failed to create decryptor: Unable to decrypt cipher INVALID"),
        ]

    @patch("streamlink.stream.hls.hls.log")
    def test_hls_encrypted_missing_uri(self, mock_log: Mock):
        aesKey, aesIv, key = self.gen_key(uri=False)

        self.subject([
            Playlist(0, [key, SegmentEnc(1, aesKey, aesIv)], end=True),
        ])
        self.await_write()

        self.thread.close()
        self.await_close()

        assert b"".join(self.thread.data) == b""
        assert mock_log.error.mock_calls == [
            call("Failed to create decryptor: Missing URI for decryption key"),
        ]

    @patch("streamlink.stream.hls.hls.log")
    def test_hls_encrypted_missing_adapter(self, mock_log: Mock):
        aesKey, aesIv, key = self.gen_key(uri="foo://bar/baz", mock={"exc": InvalidSchema})

        self.subject([
            Playlist(0, [key, SegmentEnc(1, aesKey, aesIv)], end=True),
        ])
        self.await_write()

        self.thread.close()
        self.await_close()

        assert b"".join(self.thread.data) == b""
        assert mock_log.error.mock_calls == [
            call("Failed to create decryptor: Unable to find connection adapter for key URI: foo://bar/baz"),
        ]

    def test_hls_encrypted_aes128(self):
        aesKey, aesIv, key = self.gen_key()
        long = b"Test cipher block chaining mode by using a long bytes string"

        # noinspection PyTypeChecker
        segments = self.subject([
            Playlist(0, [key] + [SegmentEnc(num, aesKey, aesIv) for num in range(4)]),
            Playlist(4, [key] + [SegmentEnc(num, aesKey, aesIv, content=long) for num in range(4, 8)], end=True),
        ])

        self.await_write(3 + 4)
        data = self.await_read(read_all=True)
        self.await_close()

        expected = self.content(segments, prop="content_plain", cond=lambda s: s.num >= 1)
        assert data == expected, "Decrypts the AES-128 identity stream"
        assert self.called(key, once=True), "Downloads encryption key only once"
        assert self.get_mock(key).last_request._request.headers.get("X-FOO") == "BAR"
        assert not any(self.called(s) for s in segments.values() if s.num < 1), "Skips first segment"
        assert all(self.called(s) for s in segments.values() if s.num >= 1), "Downloads all remaining segments"
        assert self.get_mock(segments[1]).last_request._request.headers.get("X-FOO") == "BAR"

    def test_hls_encrypted_aes128_with_map(self):
        aesKey, aesIv, key = self.gen_key()
        map1 = TagMapEnc(1, namespace=self.id(), key=aesKey, iv=aesIv)
        map2 = TagMapEnc(2, namespace=self.id(), key=aesKey, iv=aesIv)
        self.mock("GET", self.url(map1), content=map1.content)
        self.mock("GET", self.url(map2), content=map2.content)

        segments = self.subject([
            Playlist(0, [key, map1] + [SegmentEnc(num, aesKey, aesIv) for num in range(2)]),
            Playlist(2, [key, map2] + [SegmentEnc(num, aesKey, aesIv) for num in range(2, 4)], end=True),
        ])

        self.await_write(1 + 2 + 1 + 2)  # 1 map, 2 segments, 1 map, 2 segments
        data = self.await_read(read_all=True)
        self.await_close()

        assert data == self.content(
            [
                map1,
                segments[0],
                segments[1],
                map2,
                segments[2],
                segments[3],
            ],
            prop="content_plain",
        )

    def test_hls_encrypted_aes128_with_differently_encrypted_map(self):
        aesKey1, aesIv1, key1 = self.gen_key()  # init key
        aesKey2, aesIv2, key2 = self.gen_key()  # media key
        map1 = TagMapEnc(1, namespace=self.id(), key=aesKey1, iv=aesIv1)
        map2 = TagMapEnc(2, namespace=self.id(), key=aesKey1, iv=aesIv1)
        self.mock("GET", self.url(map1), content=map1.content)
        self.mock("GET", self.url(map2), content=map2.content)

        segments = self.subject([
            Playlist(0, [key1, map1, key2] + [SegmentEnc(num, aesKey2, aesIv2) for num in range(2)]),
            Playlist(2, [key1, map2, key2] + [SegmentEnc(num, aesKey2, aesIv2) for num in range(2, 4)], end=True),
        ])

        self.await_write(1 + 2 + 1 + 2)  # 1 map, 2 segments, 1 map, 2 segments
        data = self.await_read(read_all=True)
        self.await_close()

        assert data == self.content(
            [
                map1,
                segments[0],
                segments[1],
                map2,
                segments[2],
                segments[3],
            ],
            prop="content_plain",
        )

    def test_hls_encrypted_aes128_with_plaintext_map(self):
        aesKey, aesIv, key = self.gen_key()
        map1 = TagMap(1, namespace=self.id())
        map2 = TagMap(2, namespace=self.id())
        self.mock("GET", self.url(map1), content=map1.content)
        self.mock("GET", self.url(map2), content=map2.content)

        segments = self.subject([
            Playlist(0, [map1, key] + [SegmentEnc(num, aesKey, aesIv) for num in range(2)]),
            Playlist(2, [map2, key] + [SegmentEnc(num, aesKey, aesIv) for num in range(2, 4)], end=True),
        ])

        self.await_write(1 + 2 + 1 + 2)  # 1 map, 2 segments, 1 map, 2 segments
        data = self.await_read(read_all=True)
        self.await_close()

        assert data == (
            map1.content
            + segments[0].content_plain
            + segments[1].content_plain
            + map2.content
            + segments[2].content_plain
            + segments[3].content_plain
        )

    def test_hls_encrypted_aes128_key_uri_override(self):
        aesKey, aesIv, key = self.gen_key(uri="http://real-mocked/{namespace}/encryption.key?foo=bar")
        aesKeyInvalid = bytes(ord(aesKey[i : i + 1]) ^ 0xFF for i in range(16))
        _, __, key_invalid = self.gen_key(aesKeyInvalid, aesIv, uri="http://mocked/{namespace}/encryption.key?foo=bar")

        # noinspection PyTypeChecker
        segments = self.subject(
            [
                Playlist(0, [key_invalid] + [SegmentEnc(num, aesKey, aesIv) for num in range(4)]),
                Playlist(4, [key_invalid] + [SegmentEnc(num, aesKey, aesIv) for num in range(4, 8)], end=True),
            ],
            options={"hls-segment-key-uri": "{scheme}://real-{netloc}{path}?{query}"},
        )

        self.await_write(3 + 4)
        data = self.await_read(read_all=True)
        self.await_close()

        expected = self.content(segments, prop="content_plain", cond=lambda s: s.num >= 1)
        assert data == expected, "Decrypts stream from custom key"
        assert not self.called(key_invalid), "Skips encryption key"
        assert self.called(key, once=True), "Downloads custom encryption key"
        assert self.get_mock(key).last_request._request.headers.get("X-FOO") == "BAR"

    @patch("streamlink.stream.hls.hls.log")
    def test_hls_encrypted_aes128_incorrect_block_length(self, mock_log: Mock):
        aesKey, aesIv, key = self.gen_key()

        segments = self.subject([
            Playlist(
                0,
                [
                    key,
                    SegmentEnc(0, aesKey, aesIv, append=b"?"),
                    SegmentEnc(1, aesKey, aesIv),
                ],
                end=True,
            ),
        ])
        self.await_write()
        assert self.thread.reader.writer.is_alive()

        self.await_write()
        data = self.await_read(read_all=True)
        self.await_close()

        assert data == self.content([segments[1]], prop="content_plain")
        assert mock_log.error.mock_calls == [
            call("Error while decrypting segment 0: Data must be padded to 16 byte boundary in CBC mode"),
        ]

    @patch("streamlink.stream.hls.hls.log")
    def test_hls_encrypted_aes128_incorrect_padding_length(self, mock_log: Mock):
        aesKey, aesIv, key = self.gen_key()

        padding = b"\x00" * (AES.block_size - len(b"[0]"))
        segments = self.subject([
            Playlist(
                0,
                [
                    key,
                    SegmentEnc(0, aesKey, aesIv, padding=padding),
                    SegmentEnc(1, aesKey, aesIv),
                ],
                end=True,
            ),
        ])
        self.await_write()
        assert self.thread.reader.writer.is_alive()

        self.await_write()
        data = self.await_read(read_all=True)
        self.await_close()

        assert data == self.content([segments[1]], prop="content_plain")
        assert mock_log.error.mock_calls == [call("Error while decrypting segment 0: Padding is incorrect.")]

    @patch("streamlink.stream.hls.hls.log")
    def test_hls_encrypted_aes128_incorrect_padding_content(self, mock_log: Mock):
        aesKey, aesIv, key = self.gen_key()

        padding = (b"\x00" * (AES.block_size - len(b"[0]") - 1)) + bytes([AES.block_size])
        segments = self.subject([
            Playlist(
                0,
                [
                    key,
                    SegmentEnc(0, aesKey, aesIv, padding=padding),
                    SegmentEnc(1, aesKey, aesIv),
                ],
                end=True,
            ),
        ])
        self.await_write()
        assert self.thread.reader.writer.is_alive()

        self.await_write()
        data = self.await_read(read_all=True)
        self.await_close()

        assert data == self.content([segments[1]], prop="content_plain")
        assert mock_log.error.mock_calls == [call("Error while decrypting segment 0: PKCS#7 padding is incorrect.")]

    def test_hls_encrypted_switch_methods(self):
        aesKey1, aesIv1, key_aes128_1 = self.gen_key()
        aesKey2, aesIv2, key_aes128_2 = self.gen_key()
        key_none = Tag("EXT-X-KEY", {"METHOD": "NONE"})

        segments = self.subject([
            Playlist(
                0,
                [
                    key_aes128_1,
                    SegmentEnc(0, aesKey1, aesIv1),
                    SegmentEnc(1, aesKey1, aesIv1),
                    key_none,
                    Segment(2),
                    Segment(3),
                    key_aes128_2,
                    SegmentEnc(4, aesKey2, aesIv2),
                    SegmentEnc(5, aesKey2, aesIv2),
                ],
                end=True,
            ),
        ])

        self.await_write(6)
        data = self.await_read(read_all=True)
        self.await_close()

        expected = self.content(segments, prop="content_plain", cond=lambda s: 0 <= s.num <= 1)
        expected += self.content(segments, cond=lambda s: 2 <= s.num <= 3)
        expected += self.content(segments, prop="content_plain", cond=lambda s: 4 <= s.num <= 5)
        assert data == expected, "Switches between encryption key methods"


@patch("streamlink.stream.hls.hls.HLSStreamWorker.wait", Mock(return_value=True))
class TestHlsPlaylistReloadTime(TestMixinStreamHLS, unittest.TestCase):
    segments = [
        Segment(0, duration=11),
        Segment(1, duration=7),
        Segment(2, duration=5),
        Segment(3, duration=3),
    ]

    def get_session(self, options=None, reload_time=None, *args, **kwargs):
        return super().get_session(
            dict(
                options or {},
                **{
                    "hls-live-edge": 3,
                    "hls-playlist-reload-time": reload_time,
                },
            ),
        )

    def subject(self, *args, **kwargs):
        super().subject(*args, start=False, **kwargs)

        # mock the worker thread's _playlist_reload_time method, so that the main thread can wait on its call
        playlist_reload_time_called = Event()
        orig_playlist_reload_time = self.thread.reader.worker._playlist_reload_time

        def mocked_playlist_reload_time(*args, **kwargs):
            playlist_reload_time_called.set()
            return orig_playlist_reload_time(*args, **kwargs)

        # immediately kill the writer thread as we don't need it and don't want to wait for its queue polling to end
        def mocked_queue_get():
            return None

        with (
            patch.object(self.thread.reader.worker, "_playlist_reload_time", side_effect=mocked_playlist_reload_time),
            patch.object(self.thread.reader.writer, "_queue_get", side_effect=mocked_queue_get),
        ):
            self.start()

            if not playlist_reload_time_called.wait(timeout=5):  # pragma: no cover
                raise RuntimeError("Missing _playlist_reload_time() call")

            # wait for the worker thread to terminate, so that deterministic assertions can be done about the reload time
            self.thread.reader.worker.join()

            return self.thread.reader.worker.playlist_reload_time

    def test_hls_playlist_reload_time_default(self):
        time = self.subject([Playlist(0, self.segments, end=True, targetduration=4)], reload_time="default")
        assert time == 4, "default sets the reload time to the playlist's target duration"

    def test_hls_playlist_reload_time_segment(self):
        time = self.subject([Playlist(0, self.segments, end=True, targetduration=4)], reload_time="segment")
        assert time == 3, "segment sets the reload time to the playlist's last segment"

    def test_hls_playlist_reload_time_segment_no_segments(self):
        time = self.subject([Playlist(0, [], end=True, targetduration=4)], reload_time="segment")
        assert time == 4, "segment sets the reload time to the targetduration if no segments are available"

    def test_hls_playlist_reload_time_segment_no_segments_no_targetduration(self):
        time = self.subject([Playlist(0, [], end=True, targetduration=0)], reload_time="segment")
        assert time == 6, "sets reload time to 6 seconds when no segments and no targetduration are available"

    def test_hls_playlist_reload_time_live_edge(self):
        time = self.subject([Playlist(0, self.segments, end=True, targetduration=4)], reload_time="live-edge")
        assert time == 8, "live-edge sets the reload time to the sum of the number of segments of the live-edge"

    def test_hls_playlist_reload_time_live_edge_no_segments(self):
        time = self.subject([Playlist(0, [], end=True, targetduration=4)], reload_time="live-edge")
        assert time == 4, "live-edge sets the reload time to the targetduration if no segments are available"

    def test_hls_playlist_reload_time_live_edge_no_segments_no_targetduration(self):
        time = self.subject([Playlist(0, [], end=True, targetduration=0)], reload_time="live-edge")
        assert time == 6, "sets reload time to 6 seconds when no segments and no targetduration are available"

    def test_hls_playlist_reload_time_number(self):
        time = self.subject([Playlist(0, self.segments, end=True, targetduration=4)], reload_time="2")
        assert time == 2, "number values override the reload time"

    def test_hls_playlist_reload_time_number_invalid(self):
        time = self.subject([Playlist(0, self.segments, end=True, targetduration=4)], reload_time="0")
        assert time == 4, "invalid number values set the reload time to the playlist's targetduration"

    def test_hls_playlist_reload_time_no_target_duration(self):
        time = self.subject([Playlist(0, self.segments, end=True, targetduration=0)], reload_time="default")
        assert time == 8, "uses the live-edge sum if the playlist is missing the targetduration data"

    def test_hls_playlist_reload_time_no_data(self):
        time = self.subject([Playlist(0, [], end=True, targetduration=0)], reload_time="default")
        assert time == 6, "sets reload time to 6 seconds when no data is available"


@patch("streamlink.stream.hls.hls.log")
@patch("streamlink.stream.hls.hls.HLSStreamWorker.wait", Mock(return_value=True))
@patch("streamlink.stream.hls.hls.HLSStreamWorker._segment_queue_timing_threshold_reached", Mock(return_value=False))
class TestHlsPlaylistParseErrors(TestMixinStreamHLS, unittest.TestCase):
    __stream__ = EventedWriterHLSStream

    class FakePlaylist(NamedTuple):
        is_master: bool = False
        iframes_only: bool = False

    class InvalidPlaylist(Playlist):
        def build(self, *args, **kwargs):
            return "invalid"

    def test_generic(self, mock_log):
        self.subject([self.InvalidPlaylist()])
        assert self.await_read(read_all=True) == b""
        self.await_close()
        assert self.thread.reader.buffer.closed, "Closes the stream on initial playlist parsing error"
        assert mock_log.debug.mock_calls == [call("Reloading playlist")]
        assert mock_log.error.mock_calls == [call("Missing #EXTM3U header")]

    def test_reload(self, mock_log):
        segments = self.subject([
            Playlist(1, [Segment(0)]),
            self.InvalidPlaylist(),
            self.InvalidPlaylist(),
            Playlist(2, [Segment(2)], end=True),
        ])
        self.await_write(2)
        data = self.await_read(read_all=True)
        assert data == self.content(segments)
        self.close()
        self.await_close()
        assert mock_log.warning.mock_calls == [
            call("Failed to reload playlist: Missing #EXTM3U header"),
            call("Failed to reload playlist: Missing #EXTM3U header"),
        ]

    @patch("streamlink.stream.hls.hls.parse_m3u8", Mock(return_value=FakePlaylist(is_master=True)))
    def test_is_master(self, mock_log):
        self.subject([Playlist()])
        assert self.await_read(read_all=True) == b""
        self.await_close()
        assert self.thread.reader.buffer.closed, "Closes the stream on initial playlist parsing error"
        assert mock_log.debug.mock_calls == [call("Reloading playlist")]
        assert mock_log.error.mock_calls == [
            call(f"Attempted to play a variant playlist, use 'hls://{self.stream.url}' instead"),
        ]

    @patch("streamlink.stream.hls.hls.parse_m3u8", Mock(return_value=FakePlaylist(iframes_only=True)))
    def test_iframes_only(self, mock_log):
        self.subject([Playlist()])
        assert self.await_read(read_all=True) == b""
        self.await_close()
        assert self.thread.reader.buffer.closed, "Closes the stream on initial playlist parsing error"
        assert mock_log.debug.mock_calls == [call("Reloading playlist")]
        assert mock_log.error.mock_calls == [call("Streams containing I-frames only are not playable")]


class TestHlsExtAudio:
    @pytest.fixture(autouse=True)
    def _is_usable(self, monkeypatch: pytest.MonkeyPatch):
        monkeypatch.setattr("streamlink.stream.hls.hls.FFMPEGMuxer.is_usable", Mock(return_value=True))

    @pytest.fixture(autouse=True)
    def _playlist(self, request: pytest.FixtureRequest, requests_mock: rm.Mocker):
        params = getattr(request, "param", {})
        with text(params.get("playlist", "hls/test_media_language.m3u8")) as playlist:
            requests_mock.get("http://mocked/path/master.m3u8", text=playlist.read())
            yield

    @pytest.fixture()
    def stream(self, session: Streamlink):
        streams = HLSStream.parse_variant_playlist(session, "http://mocked/path/master.m3u8")
        assert "video" in streams

        return streams["video"]

    def test_no_selection(self, stream: HLSStream):
        assert not isinstance(stream, MuxedHLSStream)
        assert stream.url == "http://mocked/path/playlist.m3u8"

    @pytest.mark.parametrize(
        ("session", "selection"),
        [
            pytest.param({"hls-audio-select": ["EN"]}, "http://mocked/path/en.m3u8", id="English-alpha2"),
            pytest.param({"hls-audio-select": ["ENG"]}, "http://mocked/path/en.m3u8", id="English-alpha3"),
            pytest.param({"hls-audio-select": ["English"]}, "http://mocked/path/en.m3u8", id="English-name"),
            pytest.param({"hls-audio-select": ["English Language"]}, "http://mocked/path/en.m3u8", id="English-name-attr"),
            pytest.param({"hls-audio-select": ["es"]}, "http://mocked/path/es.m3u8", id="Spanish-alpha2"),
            pytest.param({"hls-audio-select": ["spa"]}, "http://mocked/path/es.m3u8", id="Spanish-alpha3"),
            pytest.param({"hls-audio-select": ["spanish"]}, "http://mocked/path/es.m3u8", id="Spanish-name"),
            pytest.param({"hls-audio-select": ["spanish language"]}, "http://mocked/path/es.m3u8", id="Spanish-name-attr"),
        ],
        indirect=["session"],
    )
    def test_selection(self, session: Streamlink, stream: MuxedHLSStream, selection: str):
        assert isinstance(stream, MuxedHLSStream)
        assert [substream.url for substream in stream.substreams] == [
            "http://mocked/path/playlist.m3u8",
            selection,
        ]

    @pytest.mark.parametrize(
        "session",
        [
            pytest.param({"hls-audio-select": ["en", "es"]}, id="alpha2"),
            pytest.param({"hls-audio-select": ["eng", "spa"]}, id="alpha3"),
            pytest.param({"hls-audio-select": ["English", "Spanish"]}, id="name"),
            pytest.param({"hls-audio-select": ["English language", "Spanish language"]}, id="name-attr"),
        ],
        indirect=["session"],
    )
    def test_multiple(self, session: Streamlink, stream: MuxedHLSStream):
        assert isinstance(stream, MuxedHLSStream)
        assert [substream.url for substream in stream.substreams] == [
            "http://mocked/path/playlist.m3u8",
            "http://mocked/path/en.m3u8",
            "http://mocked/path/es.m3u8",
        ]

    @pytest.mark.parametrize(
        "session",
        [
            pytest.param({"hls-audio-select": ["*"]}, id="wildcard"),
        ],
        indirect=["session"],
    )
    def test_wildcard(self, session: Streamlink, stream: MuxedHLSStream):
        assert isinstance(stream, MuxedHLSStream)
        assert [substream.url for substream in stream.substreams] == [
            "http://mocked/path/playlist.m3u8",
            "http://mocked/path/en.m3u8",
            "http://mocked/path/es.m3u8",
            "http://mocked/path/de.m3u8",
        ]

    @pytest.mark.parametrize(
        ("session", "_playlist"),
        [
            pytest.param(
                {"hls-audio-select": ["und", "qaa", "invalid"]},
                {"playlist": "hls/test_media_language_special.m3u8"},
                id="special-reserved-invalid",
            ),
            pytest.param(
                {"hls-audio-select": ["Undetermined", "Reserved", "does NOT exist"]},
                {"playlist": "hls/test_media_language_special.m3u8"},
                id="name-attribute",
            ),
        ],
        indirect=["session", "_playlist"],
    )
    def test_parse_media_language(self, caplog: pytest.LogCaptureFixture, session: Streamlink, stream: MuxedHLSStream):
        assert isinstance(stream, MuxedHLSStream)
        assert [substream.url for substream in stream.substreams] == [
            "http://mocked/path/playlist.m3u8",
            "http://mocked/path/qaa.m3u8",
            "http://mocked/path/invalid.m3u8",
            "http://mocked/path/und.m3u8",
        ]
        assert [
            record.message
            for record in caplog.get_records(when="setup")
            if record.name == "streamlink.stream.hls" and record.levelno == logging.WARNING
        ] == [
            "Unrecognized language for media playlist: language='invalid' name='Does not exist'",
        ]

    @pytest.mark.parametrize(
        "session",
        [
            pytest.param({"hls-audio-select": ["*"]}, id="names"),
        ],
        indirect=["session"],
    )
    def test_substream_names(self, session: Streamlink, stream: MuxedHLSStream):
        assert isinstance(stream, MuxedHLSStream)
        assert [substream.name for substream in stream.substreams] == [
            None,
            "audio",
            "audio",
            "audio",
        ]


class TestM3U8ParserLogging:
    @pytest.mark.parametrize(("loglevel", "has_logs"), [("trace", False), ("all", True)])
    def test_log(self, caplog: pytest.LogCaptureFixture, loglevel: str, has_logs: bool):
        caplog.set_level(loglevel, "streamlink")

        parser = M3U8Parser[M3U8, HLSSegment, HLSPlaylist]()
        with text("hls/test_1.m3u8") as pl:
            data = pl.read()
        parser.parse(data)

        assert bool(caplog.records) is has_logs



import pytest

from streamlink.options import Options
from streamlink.plugins.soop import Soop
from streamlink.session import Streamlink
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlSoop(PluginCanHandleUrl):
    __plugin__ = Soop

    should_match_groups = [
        ("https://play.sooplive.co.kr/CHANNEL", {"channel": "CHANNEL"}),
        ("https://play.sooplive.co.kr/CHANNEL/0123456789", {"channel": "CHANNEL", "bno": "0123456789"}),
        ("https://play.afreecatv.com/CHANNEL", {"channel": "CHANNEL"}),
        ("https://play.afreecatv.com/CHANNEL/0123456789", {"channel": "CHANNEL", "bno": "0123456789"}),
    ]

    should_not_match = [
        "https://sooplive.co.kr/CHANNEL",
        "https://sooplive.co.kr/CHANNEL/0123456789",
        "https://afreecatv.com/CHANNEL",
        "https://afreecatv.com/CHANNEL/0123456789",
    ]


@pytest.mark.parametrize(
    ("plugin_options", "expected"),
    [
        pytest.param(
            {"username": "username", "password": "password", "purge-credentials": True, "stream-password": "foo"},
            {"username": "username", "password": "password", "purge-credentials": True, "stream-password": "foo"},
            id="regular-options",
        ),
        pytest.param(
            {
                "afreeca-username": "username",
                "afreeca-password": "password",
                "afreeca-purge-credentials": True,
                "afreeca-stream-password": "foo",
            },
            {
                "username": "username",
                "password": "password",
                "purge-credentials": True,
                "stream-password": "foo",
                "afreeca-username": "username",
                "afreeca-password": "password",
                "afreeca-purge-credentials": True,
                "afreeca-stream-password": "foo",
            },
            id="deprecated-options",
        ),
        pytest.param(
            {
                "username": "username1",
                "password": "password1",
                "purge-credentials": True,
                "stream-password": "foo",
                "afreeca-username": "username2",
                "afreeca-password": "password2",
                "afreeca-purge-credentials": False,
                "afreeca-stream-password": "bar",
            },
            {
                "username": "username1",
                "password": "password1",
                "purge-credentials": True,
                "stream-password": "foo",
                "afreeca-username": "username2",
                "afreeca-password": "password2",
                "afreeca-purge-credentials": False,
                "afreeca-stream-password": "bar",
            },
            id="regular-and-deprecated-options",
        ),
    ],
)
def test_options(session: Streamlink, plugin_options: dict, expected: dict):
    options = Options()
    options.update(plugin_options)
    plugin = Soop(session, "https://play.sooplive.co.kr/CHANNEL/0123456789", options)
    assert dict(plugin.options.items()) == expected



from streamlink.plugins.turkuvaz import Turkuvaz
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlTurkuvaz(PluginCanHandleUrl):
    __plugin__ = Turkuvaz

    should_match_groups = [
        # canonical live links
        (("ahaber", "https://www.ahaber.com.tr/video/canli-yayin"), {}),  # ahbr
        (("ahaber", "https://www.ahaber.com.tr/canli-yayin-aspor.html"), {}),  # ahbr: aspor
        (("ahaber", "https://www.ahaber.com.tr/canli-yayin-anews.html"), {}),  # ahbr: anews
        (("ahaber", "https://www.ahaber.com.tr/canli-yayin-a2tv.html"), {}),  # ahbr: a2tv
        (("ahaber", "https://www.ahaber.com.tr/canli-yayin-minikago.html"), {}),  # ahbr: minika go
        (("ahaber", "https://www.ahaber.com.tr/canli-yayin-minikacocuk.html"), {}),  # ahbr: minika cocuk
        (("anews", "https://www.anews.com.tr/webtv/live-broadcast"), {}),  # anews
        (("apara", "https://www.apara.com.tr/canli-yayin"), {}),  # apara
        (("aspor", "https://www.aspor.com.tr/webtv/canli-yayin"), {}),  # aspor
        (("atv", "https://www.atv.com.tr/canli-yayin"), {}),  # atv
        (("atv", "https://www.atv.com.tr/a2tv/canli-yayin"), {}),  # a2tv
        (("atvavrupa", "https://www.atvavrupa.tv/canli-yayin"), {}),  # atvavrupa
        (("minikacocuk", "https://www.minikacocuk.com.tr/webtv/canli-yayin"), {}),  # minika cocuk
        (("minikago", "https://www.minikago.com.tr/webtv/canli-yayin"), {}),  # minika go
        (("vavtv", "https://vavtv.com.tr/canli-yayin"), {}),  # vavtv
        # vods
        (("ahaber", "https://www.ahaber.com.tr/video/yasam-videolari/samsunda-sel-sularindan-kacma-ani-kamerada"), {}),
        (("anews", "https://www.anews.com.tr/webtv/world/pro-ukraine-militia-says-it-has-captured-russian-soldiers"), {}),
        (("apara", "https://www.apara.com.tr/video/ekonomide-bugun/bist-100de-kar-satislari-derinlesir-mi"), {}),
        (("aspor", "https://www.aspor.com.tr/webtv/galatasaray/galatasaraydan-forma-tanitiminda-fenerbahceye-gonderme"), {}),
        (("atvavrupa", "https://www.atvavrupa.tv/diger-program//ozelvideo/izle"), {}),
        (("minikago", "https://www.minikago.com.tr/webtv/mondo-yan/bolum/mondo-yan-eylul-tanitim"), {}),
        (("vavtv", "https://vavtv.com.tr/vavradyo/video/guncel/munafiklarin-ozellikleri-nelerdir"), {}),
        # other links for info/doc
        (("atv", "https://www.atv.com.tr/webtv/canli-yayin"), {}),  # redirect to atv.com.tr/canli-yayin
        (("ahaber", "https://www.ahaber.com.tr/canli-yayin-atv.html"), {}),  # links to atv.com.tr/webtv/canli-yayin
    ]



from __future__ import annotations

import argparse
import logging
import re
import time
from contextlib import nullcontext
from operator import eq, gt, lt
from typing import Any
from unittest.mock import Mock, call, patch

import freezegun
import pytest
import requests.cookies

from streamlink.options import Options
from streamlink.plugin import (
    HIGH_PRIORITY,
    NORMAL_PRIORITY,
    Plugin,
    PluginArgument,
    PluginArguments,
    PluginError,
    pluginargument,
    pluginmatcher,
)

# noinspection PyProtectedMember
from streamlink.plugin.plugin import (
    _COOKIE_KEYS,  # noqa: PLC2701
    _PLUGINARGUMENT_TYPE_REGISTRY,  # noqa: PLC2701
    Matcher,
    parse_params,
    stream_weight,
)
from streamlink.session import Streamlink


class FakePlugin(Plugin):
    def _get_streams(self):
        pass  # pragma: no cover


class RenamedPlugin(FakePlugin):
    __module__ = "foo.bar.baz"


class CustomConstructorOnePlugin(FakePlugin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class CustomConstructorTwoPlugin(FakePlugin):
    def __init__(self, session, url):
        super().__init__(session, url)


class TestPlugin:
    @pytest.mark.parametrize(
        ("pluginclass", "module", "logger"),
        [
            (Plugin, "plugin", "streamlink.plugin.plugin"),
            (FakePlugin, "test_plugin", "tests.test_plugin"),
            (RenamedPlugin, "baz", "foo.bar.baz"),
            (CustomConstructorOnePlugin, "test_plugin", "tests.test_plugin"),
            (CustomConstructorTwoPlugin, "test_plugin", "tests.test_plugin"),
        ],
    )
    def test_constructor(
        self,
        monkeypatch: pytest.MonkeyPatch,
        caplog: pytest.LogCaptureFixture,
        pluginclass: type[Plugin],
        module: str,
        logger: str,
        session: Streamlink,
    ):
        mock_cache = Mock()
        mock_load_cookies = Mock()
        monkeypatch.setattr("streamlink.plugin.plugin.Cache", mock_cache)
        monkeypatch.setattr(pluginclass, "load_cookies", mock_load_cookies)

        plugin = pluginclass(session, "http://localhost")

        assert not caplog.records

        assert plugin.session is session
        assert plugin.url == "http://localhost"

        assert plugin.module == module

        assert isinstance(plugin.logger, logging.Logger)
        assert plugin.logger.name == logger

        assert mock_cache.call_args_list == [call(filename="plugin-cache.json", key_prefix=module, disabled=False)]
        assert plugin.cache == mock_cache()

        assert mock_load_cookies.call_args_list == [call()]

    def test_constructor_options(self):
        one = FakePlugin(Mock(), "https://mocked", Options({"key": "val"}))
        two = FakePlugin(Mock(), "https://mocked")
        assert one.get_option("key") == "val"
        assert two.get_option("key") is None

        one.set_option("key", "other")
        assert one.get_option("key") == "other"
        assert two.get_option("key") is None

    @pytest.mark.parametrize(
        ("session", "expected"),
        [
            pytest.param({}, False, id="default"),
            pytest.param({"no-plugin-cache": False}, False, id="not-disabled"),
            pytest.param({"no-plugin-cache": True}, True, id="disabled"),
        ],
        indirect=["session"],
    )
    def test_disabled_cache(self, session: Streamlink, expected: bool):
        plugin = FakePlugin(session, "https://mocked")
        assert plugin.cache._disabled is expected


class TestPluginMatcher:
    # noinspection PyUnusedLocal
    def test_decorator(self):
        class MyPlugin:
            pass

        with pytest.raises(TypeError) as cm:
            # noinspection PyTypeChecker
            pluginmatcher(re.compile(r""))(MyPlugin)

        assert str(cm.value) == "MyPlugin is not a Plugin"

    def test_named_duplicate(self):
        class MyPlugin(FakePlugin):
            pass

        matcher = pluginmatcher(re.compile(r"http://foo"), name="foo")

        with pytest.raises(ValueError, match=r"^A matcher named 'foo' has already been registered$"):
            matcher(matcher(MyPlugin))

    def test_no_matchers(self):
        class MyPlugin(FakePlugin):
            pass

        plugin = MyPlugin(Mock(), "http://foo")
        assert plugin.url == "http://foo"
        assert plugin.matchers == []
        assert plugin.matches == []
        assert plugin.matcher is None
        assert plugin.match is None

    def test_matchers(self):
        @pluginmatcher(re.compile(r"foo", re.VERBOSE))
        @pluginmatcher(re.compile(r"bar"), priority=HIGH_PRIORITY)
        @pluginmatcher(re.compile(r"baz"), priority=HIGH_PRIORITY, name="baz")
        class MyPlugin(FakePlugin):
            pass

        assert MyPlugin.matchers == [
            Matcher(re.compile(r"foo", re.VERBOSE), NORMAL_PRIORITY),
            Matcher(re.compile(r"bar"), HIGH_PRIORITY),
            Matcher(re.compile(r"baz"), HIGH_PRIORITY, "baz"),
        ]

    def test_matchers_inheritance(self):
        @pluginmatcher(re.compile(r"foo"))
        @pluginmatcher(re.compile(r"bar"))
        class PluginOne(FakePlugin):
            pass

        @pluginmatcher(re.compile(r"baz"))
        @pluginmatcher(re.compile(r"qux"))
        class PluginTwo(PluginOne):
            pass

        assert PluginOne.matchers is not PluginTwo.matchers
        assert PluginOne.matchers == [
            Matcher(re.compile(r"foo"), NORMAL_PRIORITY),
            Matcher(re.compile(r"bar"), NORMAL_PRIORITY),
        ]
        assert PluginTwo.matchers == [
            Matcher(re.compile(r"baz"), NORMAL_PRIORITY),
            Matcher(re.compile(r"qux"), NORMAL_PRIORITY),
            Matcher(re.compile(r"foo"), NORMAL_PRIORITY),
            Matcher(re.compile(r"bar"), NORMAL_PRIORITY),
        ]

    # noinspection PyUnusedLocal
    def test_matchers_inheritance_named_duplicate(self):
        @pluginmatcher(name="foo", pattern=re.compile(r"foo"))
        class PluginOne(FakePlugin):
            pass

        with pytest.raises(ValueError, match=r"^A matcher named 'foo' has already been registered$"):

            @pluginmatcher(name="foo", pattern=re.compile(r"foo"))
            class PluginTwo(PluginOne):
                pass

    def test_matchers_not_matching(self):
        @pluginmatcher(re.compile(r"http://foo"))
        class MyPlugin(FakePlugin):
            pass

        with pytest.raises(PluginError, match=r"^The input URL did not match any of this plugin's matchers$"):
            MyPlugin(Mock(), "http://bar")

    def test_no_matchers_not_matching(self):
        plugin = FakePlugin(Mock(), "")
        assert plugin.match is None

    def test_url_setter(self):
        @pluginmatcher(re.compile(r"http://(foo)"))
        @pluginmatcher(re.compile(r"http://(bar)"))
        @pluginmatcher(re.compile(r"http://(baz)"))
        class MyPlugin(FakePlugin):
            pass

        plugin = MyPlugin(Mock(), "http://foo")
        assert plugin.url == "http://foo"
        assert [m is not None for m in plugin.matches] == [True, False, False]
        assert plugin.matcher is plugin.matchers[0].pattern
        assert plugin.match.group(1) == "foo"

        plugin.url = "http://bar"
        assert plugin.url == "http://bar"
        assert [m is not None for m in plugin.matches] == [False, True, False]
        assert plugin.matcher is plugin.matchers[1].pattern
        assert plugin.match.group(1) == "bar"

        plugin.url = "http://baz"
        assert plugin.url == "http://baz"
        assert [m is not None for m in plugin.matches] == [False, False, True]
        assert plugin.matcher is plugin.matchers[2].pattern
        assert plugin.match.group(1) == "baz"

        plugin.url = "http://qux"
        assert plugin.url == "http://qux"
        assert [m is not None for m in plugin.matches] == [False, False, False]
        assert plugin.matcher is None
        assert plugin.match is None

    def test_named_matchers_and_matches(self):
        @pluginmatcher(re.compile(r"http://foo"), name="foo")
        @pluginmatcher(re.compile(r"http://bar"), name="bar")
        class MyPlugin(FakePlugin):
            pass

        plugin = MyPlugin(Mock(), "http://foo")

        assert plugin.matchers["foo"] is plugin.matchers[0]
        assert plugin.matchers["bar"] is plugin.matchers[1]
        with pytest.raises(IndexError):
            plugin.matchers.__getitem__(2)
        with pytest.raises(KeyError):
            plugin.matchers.__getitem__("baz")

        assert plugin.matches["foo"] is plugin.matches[0]
        assert plugin.matches["bar"] is plugin.matches[1]
        assert plugin.matches["foo"] is not None
        assert plugin.matches["bar"] is None
        with pytest.raises(IndexError):
            plugin.matches.__getitem__(2)
        with pytest.raises(KeyError):
            plugin.matches.__getitem__("baz")

        plugin.url = "http://bar"
        assert plugin.matches["foo"] is None
        assert plugin.matches["bar"] is not None

        plugin.url = "http://baz"
        assert plugin.matches["foo"] is None
        assert plugin.matches["bar"] is None


class TestPluginArguments:
    @pluginargument("foo", dest="_foo", help="FOO")
    @pluginargument("bar", dest="_bar", help="BAR")
    @pluginargument("baz", dest="_baz", help="BAZ")
    class DecoratedPlugin(FakePlugin):
        pass

    class ClassAttrPlugin(FakePlugin):
        arguments = PluginArguments(
            PluginArgument("foo", dest="_foo", help="FOO"),
            PluginArgument("bar", dest="_bar", help="BAR"),
            PluginArgument("baz", dest="_baz", help="BAZ"),
        )

    def test_pluginargument_type_registry(self):
        assert _PLUGINARGUMENT_TYPE_REGISTRY
        assert all(callable(value) for value in _PLUGINARGUMENT_TYPE_REGISTRY.values())

    @pytest.mark.parametrize("pluginclass", [DecoratedPlugin, ClassAttrPlugin])
    def test_arguments(self, pluginclass):
        assert pluginclass.arguments is not None
        assert tuple(arg.name for arg in pluginclass.arguments) == ("foo", "bar", "baz"), "Argument name"
        assert tuple(arg.dest for arg in pluginclass.arguments) == ("_foo", "_bar", "_baz"), "Argument keyword"
        assert tuple(arg.options.get("help") for arg in pluginclass.arguments) == ("FOO", "BAR", "BAZ"), "argparse keyword"

    @pytest.mark.parametrize("pluginclass", [DecoratedPlugin, ClassAttrPlugin])
    def test_arguments_mixed(self, pluginclass):
        @pluginargument("qux")
        class MixedPlugin(pluginclass):
            pass

        assert tuple(arg.name for arg in MixedPlugin.arguments) == ("qux", "foo", "bar", "baz")

    def test_arguments_inheritance(self):
        @pluginargument("foo", help="FOO")
        @pluginargument("bar", help="BAR")
        class PluginOne(FakePlugin):
            pass

        @pluginargument("baz", help="BAZ")
        @pluginargument("qux", help="QUX")
        class PluginTwo(PluginOne):
            pass

        assert PluginOne.arguments is not PluginTwo.arguments
        assert tuple(arg.name for arg in PluginOne.arguments) == ("foo", "bar")
        assert tuple(arg.name for arg in PluginTwo.arguments) == ("baz", "qux", "foo", "bar")

    @pytest.mark.parametrize(
        ("options", "args", "expected", "raises"),
        [
            pytest.param(
                {"type": "int"},
                ["--myplugin-foo", "123"],
                123,
                nullcontext(),
                id="int",
            ),
            pytest.param(
                {"type": "float"},
                ["--myplugin-foo", "123.456"],
                123.456,
                nullcontext(),
                id="float",
            ),
            pytest.param(
                {"type": "bool"},
                ["--myplugin-foo", "yes"],
                True,
                nullcontext(),
                id="bool",
            ),
            pytest.param(
                {"type": "keyvalue"},
                ["--myplugin-foo", "key=value"],
                ("key", "value"),
                nullcontext(),
                id="keyvalue",
            ),
            pytest.param(
                {"type": "comma_list_filter", "type_args": (["one", "two", "four"],)},
                ["--myplugin-foo", "one,two,three,four"],
                ["one", "two", "four"],
                nullcontext(),
                id="comma_list_filter - args",
            ),
            pytest.param(
                {"type": "comma_list_filter", "type_kwargs": {"acceptable": ["one", "two", "four"]}},
                ["--myplugin-foo", "one,two,three,four"],
                ["one", "two", "four"],
                nullcontext(),
                id="comma_list_filter - kwargs",
            ),
            pytest.param(
                {"type": "hours_minutes_seconds"},
                ["--myplugin-foo", "1h2m3s"],
                3723,
                nullcontext(),
                id="hours_minutes_seconds",
            ),
            pytest.param(
                {"type": "UNKNOWN"},
                None,
                None,
                pytest.raises(TypeError),
                id="UNKNOWN",
            ),
        ],
    )
    def test_type_argument_map(self, options: dict, args: list, expected: Any, raises: nullcontext):
        class MyPlugin(FakePlugin):
            pass

        with raises:
            pluginargument("foo", **options)(MyPlugin)
            assert MyPlugin.arguments is not None
            pluginarg = MyPlugin.arguments.get("foo")
            assert pluginarg

            parser = argparse.ArgumentParser()
            parser.add_argument(pluginarg.argument_name("myplugin"), **pluginarg.options)
            namespace = parser.parse_args(args)
            assert namespace.myplugin_foo == expected

    def test_decorator_typeerror(self, monkeypatch: pytest.MonkeyPatch):
        class NotAPluginMeta(type):
            def __repr__(self) -> str:
                return self.__name__

        with pytest.raises(TypeError, match=r"^NotAPlugin is not a Plugin$"):
            # noinspection PyUnusedLocal
            @pluginargument("foo")  # type: ignore[arg-type]
            class NotAPlugin(metaclass=NotAPluginMeta):
                pass

    def test_empty(self):
        assert FakePlugin.arguments is not None
        assert tuple(iter(FakePlugin.arguments)) == ()


@pytest.mark.parametrize("attr", ["id", "author", "category", "title"])
def test_plugin_metadata(attr):
    plugin = FakePlugin(Mock(), "https://foo.bar/")
    getter = getattr(plugin, f"get_{attr}")
    assert callable(getter)

    assert getattr(plugin, attr) is None
    assert getter() is None

    setattr(plugin, attr, " foo bar ")
    assert getter() == "foo bar"

    class Foo:
        def __str__(self):
            return " baz qux "

    setattr(plugin, attr, Foo())
    assert getter() == "baz qux"


class TestCookies:
    @staticmethod
    def create_cookie_dict(name, value, expires=None):
        return dict(
            version=0,
            name=name,
            value=value,
            port=None,
            domain="test.se",
            path="/",
            secure=False,
            expires=expires,
            discard=True,
            comment=None,
            comment_url=None,
            rest={"HttpOnly": None},
            rfc2109=False,
        )

    # TODO: py39 support end: remove explicit dummy context binding of static method
    _create_cookie_dict = create_cookie_dict.__get__(object)

    @pytest.fixture()
    def pluginclass(self):
        class MyPlugin(FakePlugin):
            __module__ = "myplugin"

        return MyPlugin

    @pytest.fixture()
    def plugincache(self, request):
        with patch("streamlink.plugin.plugin.Cache") as mock_cache:
            cache = mock_cache("plugin-cache.json", "myplugin")
            cache.get_all.return_value = request.param
            yield cache

    @pytest.fixture()
    def logger(self, pluginclass: type[Plugin]):
        with patch("streamlink.plugin.plugin.logging") as mock_logging:
            yield mock_logging.getLogger(pluginclass.__module__)

    @pytest.fixture()
    def plugin(self, pluginclass: type[Plugin], session: Streamlink, plugincache: Mock, logger: Mock):
        plugin = pluginclass(session, "http://test.se")
        assert plugin.cache is plugincache
        assert plugin.logger is logger
        return plugin

    @staticmethod
    def _cookie_to_dict(cookie):
        r = {name: getattr(cookie, name, None) for name in _COOKIE_KEYS}
        r["rest"] = getattr(cookie, "rest", getattr(cookie, "_rest", None))
        return r

    def _cookies_to_list(self, cookies):
        return [self._cookie_to_dict(cookie) for cookie in cookies]

    @pytest.mark.parametrize(
        "plugincache",
        [
            {
                "__cookie:test-name1:test.se:80:/": _create_cookie_dict("test-name1", "test-value1"),
                "__cookie:test-name2:test.se:80:/": _create_cookie_dict("test-name2", "test-value2"),
                "unrelated": "data",
            },
        ],
        indirect=True,
    )
    def test_load(self, session: Streamlink, plugin: Plugin, plugincache: Mock, logger: Mock):
        assert self._cookies_to_list(session.http.cookies) == self._cookies_to_list([
            requests.cookies.create_cookie("test-name1", "test-value1", domain="test.se"),
            requests.cookies.create_cookie("test-name2", "test-value2", domain="test.se"),
        ])
        assert logger.debug.call_args_list == [call("Restored cookies: test-name1, test-name2")]

    @pytest.mark.parametrize("plugincache", [{}], indirect=True)
    def test_save(self, session: Streamlink, plugin: Plugin, plugincache: Mock, logger: Mock):
        cookie1 = requests.cookies.create_cookie("test-name1", "test-value1", domain="test.se")
        cookie2 = requests.cookies.create_cookie("test-name2", "test-value2", domain="test.se")
        session.http.cookies.set_cookie(cookie1)
        session.http.cookies.set_cookie(cookie2)

        plugin.save_cookies(lambda cookie: cookie.name == "test-name1", default_expires=3600)
        assert plugincache.set.call_args_list == [
            call(
                "__cookie:test-name1:test.se:80:/",
                self.create_cookie_dict("test-name1", "test-value1", None),
                3600,
            ),
        ]
        assert logger.debug.call_args_list == [call("Saved cookies: test-name1")]

    @freezegun.freeze_time("1970-01-01T00:00:00Z")
    @pytest.mark.parametrize("plugincache", [{}], indirect=True)
    def test_save_expires(self, session: Streamlink, plugin: Plugin, plugincache: Mock):
        cookie = requests.cookies.create_cookie(
            "test-name",
            "test-value",
            domain="test.se",
            expires=time.time() + 3600,
            rest={"HttpOnly": None},
        )
        session.http.cookies.set_cookie(cookie)

        plugin.save_cookies(default_expires=60)
        assert plugincache.set.call_args_list == [
            call(
                "__cookie:test-name:test.se:80:/",
                self.create_cookie_dict("test-name", "test-value", 3600),
                3600,
            ),
        ]

    @pytest.mark.parametrize(
        "plugincache",
        [
            {
                "__cookie:test-name1:test.se:80:/": _create_cookie_dict("test-name1", "test-value1", None),
                "__cookie:test-name2:test.se:80:/": _create_cookie_dict("test-name2", "test-value2", None),
                "unrelated": "data",
            },
        ],
        indirect=True,
    )
    def test_clear(self, session: Streamlink, plugin: Plugin, plugincache: Mock):
        assert tuple(session.http.cookies.keys()) == ("test-name1", "test-name2")

        plugin.clear_cookies()
        assert call("__cookie:test-name1:test.se:80:/", None, 0) in plugincache.set.call_args_list
        assert call("__cookie:test-name2:test.se:80:/", None, 0) in plugincache.set.call_args_list
        assert len(session.http.cookies.keys()) == 0

    @pytest.mark.parametrize(
        "plugincache",
        [
            {
                "__cookie:test-name1:test.se:80:/": _create_cookie_dict("test-name1", "test-value1", None),
                "__cookie:test-name2:test.se:80:/": _create_cookie_dict("test-name2", "test-value2", None),
                "unrelated": "data",
            },
        ],
        indirect=True,
    )
    def test_clear_filter(self, session: Streamlink, plugin: Plugin, plugincache: Mock):
        assert tuple(session.http.cookies.keys()) == ("test-name1", "test-name2")

        plugin.clear_cookies(lambda cookie: cookie.name == "test-name2")
        assert call("__cookie:test-name1:test.se:80:/", None, 0) not in plugincache.set.call_args_list
        assert call("__cookie:test-name2:test.se:80:/", None, 0) in plugincache.set.call_args_list
        assert tuple(session.http.cookies.keys()) == ("test-name1",)


@pytest.mark.parametrize(
    ("params", "expected"),
    [
        (
            None,
            {},
        ),
        (
            "foo=bar",
            dict(foo="bar"),
        ),
        (
            "verify=False",
            dict(verify=False),
        ),
        (
            "timeout=123.45",
            dict(timeout=123.45),
        ),
        (
            "verify=False params={'key': 'a value'}",
            dict(verify=False, params=dict(key="a value")),
        ),
        (
            "\"conn=['B:1', 'S:authMe', 'O:1', 'NN:code:1.23', 'NS:flag:ok', 'O:0']",
            dict(conn=["B:1", "S:authMe", "O:1", "NN:code:1.23", "NS:flag:ok", "O:0"]),
        ),
    ],
)
def test_parse_params(params, expected):
    assert parse_params(params) == expected


@pytest.mark.parametrize(
    ("weight", "expected"),
    [
        ("720p", (720, "pixels")),
        ("720p+", (721, "pixels")),
        ("720p60", (780, "pixels")),
    ],
)
def test_stream_weight_value(weight, expected):
    assert stream_weight(weight) == expected


@pytest.mark.parametrize(
    ("weight_a", "operator", "weight_b"),
    [
        ("720p+", gt, "720p"),
        ("720p_3000k", gt, "720p_2500k"),
        ("720p60_3000k", gt, "720p_3000k"),
        ("3000k", gt, "2500k"),
        ("720p", eq, "720p"),
        ("720p_3000k", lt, "720p+_3000k"),
        # with audio
        ("720p+a256k", gt, "720p+a128k"),
        ("720p+a256k", gt, "360p+a256k"),
        ("720p+a128k", gt, "360p+a256k"),
    ],
)
def test_stream_weight(weight_a, weight_b, operator):
    assert operator(stream_weight(weight_a), stream_weight(weight_b))



import re

import pytest
import requests_mock as rm

import streamlink_cli.main
from streamlink.plugin import Plugin, pluginmatcher
from streamlink.session import Streamlink


@pytest.fixture(autouse=True)
def _plugins(session: Streamlink):
    @pluginmatcher(re.compile(r"http://exists$"))
    class FakePlugin(Plugin):
        def _get_streams(self):  # pragma: no cover
            pass

    session.plugins.update({"plugin": FakePlugin})


@pytest.mark.parametrize(
    ("argv", "exit_code"),
    [
        pytest.param(
            ["--can-handle-url", "http://aborted"],
            130,
            id="aborted",
        ),
        pytest.param(
            ["--can-handle-url", "http://exists"],
            0,
            id="exists",
        ),
        pytest.param(
            ["--can-handle-url", "http://exists-redirect"],
            0,
            id="exists-redirect",
        ),
        pytest.param(
            ["--can-handle-url", "http://missing"],
            1,
            id="missing",
        ),
        pytest.param(
            ["--can-handle-url", "http://missing-redirect"],
            1,
            id="missing-redirect",
        ),
        pytest.param(
            ["--can-handle-url-no-redirect", "http://exists"],
            0,
            id="noredirect-exists",
        ),
        pytest.param(
            ["--can-handle-url-no-redirect", "http://exists-redirect"],
            1,
            id="noredirect-exists-redirect",
        ),
        pytest.param(
            ["--can-handle-url-no-redirect", "http://missing"],
            1,
            id="noredirect-missing",
        ),
        pytest.param(
            ["--can-handle-url-no-redirect", "http://missing-redirect"],
            1,
            id="noredirect-missing-redirect",
        ),
    ],
    indirect=["argv"],
)
def test_can_handle_url(requests_mock: rm.Mocker, session: Streamlink, argv: list, exit_code: int):
    requests_mock.request(rm.ANY, "http://aborted", exc=KeyboardInterrupt)  # type: ignore[arg-type]
    requests_mock.request(rm.ANY, "http://exists", content=b"")
    requests_mock.request(rm.ANY, "http://exists-redirect", status_code=301, headers={"Location": "http://exists"})
    requests_mock.request(rm.ANY, "http://missing-redirect", status_code=301, headers={"Location": "http://missing"})

    with pytest.raises(SystemExit) as exc_info:
        streamlink_cli.main.main()
    assert exc_info.value.code == exit_code



from streamlink.plugins.ruv import Ruv
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlRuv(PluginCanHandleUrl):
    __plugin__ = Ruv

    should_match_groups = [
        (("live", "https://www.ruv.is/sjonvarp/beint/ruv"), {"channel": "ruv"}),
        (("live", "https://www.ruv.is/sjonvarp/beint/ruv2"), {"channel": "ruv2"}),
        (("live", "https://www.ruv.is/utvarp/beint/ras1"), {"channel": "ras1"}),
        (("live", "https://www.ruv.is/utvarp/beint/ras2"), {"channel": "ras2"}),
        (("vod", "https://www.ruv.is/sjonvarp/spila/vedur/30766/agcf5m"), {"id": "30766", "episode": "agcf5m"}),
        (("vod", "https://www.ruv.is/sjonvarp/spila/kastljos/35422/ahpu2g"), {"id": "35422", "episode": "ahpu2g"}),
        (("vod", "https://www.ruv.is/utvarp/spila/laesi/37102/b1qk71"), {"id": "37102", "episode": "b1qk71"}),
    ]



from __future__ import annotations

import json
from datetime import datetime, timedelta, timezone
from io import BytesIO, TextIOWrapper
from json import JSONDecodeError
from pathlib import Path
from threading import Thread, Timer
from typing import Callable
from unittest.mock import Mock

import freezegun
import pytest

# noinspection PyProtectedMember
from streamlink.cache import WRITE_DEBOUNCE_TIME, Cache
from tests.testutils.handshake import Handshake


@pytest.fixture(autouse=True)
def caplog(caplog: pytest.LogCaptureFixture):
    caplog.set_level(1, "streamlink")
    return caplog


@pytest.fixture()
def mock_load(monkeypatch: pytest.MonkeyPatch):
    mock = Mock()
    monkeypatch.setattr("streamlink.cache.Cache._load", mock)
    return mock


@pytest.fixture()
def mock_schedule_save(monkeypatch: pytest.MonkeyPatch):
    mock = Mock()
    monkeypatch.setattr("streamlink.cache.Cache._schedule_save", mock)
    return mock


@pytest.fixture()
def mock_save(monkeypatch: pytest.MonkeyPatch):
    mock = Mock()
    monkeypatch.setattr("streamlink.cache.Cache._save", mock)
    return mock


@pytest.fixture()
def no_io(mock_load: Mock, mock_schedule_save: Mock, mock_save: Mock):
    return mock_load, mock_schedule_save, mock_save


@pytest.fixture(autouse=True)
def atexit(monkeypatch: pytest.MonkeyPatch):
    stack = []

    def register(func: Callable, *args, **kwargs) -> None:
        stack.append((func, args, kwargs))

    monkeypatch.setattr("streamlink.cache._atexit_register", register)

    return stack


@pytest.fixture(autouse=True)
def mock_json_load(request: pytest.FixtureRequest, monkeypatch: pytest.MonkeyPatch):
    side_effect: Callable | Exception = getattr(request, "param", None) or json.load
    mock = Mock(side_effect=side_effect)
    monkeypatch.setattr("json.load", mock)
    return mock


@pytest.fixture(autouse=True)
def cache_dir(monkeypatch: pytest.MonkeyPatch, tmp_path: Path):
    monkeypatch.setattr("streamlink.cache.CACHE_DIR", tmp_path)
    return tmp_path


# noinspection PyProtectedMember
@pytest.fixture()
def cache(request: pytest.FixtureRequest, cache_dir: Path, atexit: list):
    param = getattr(request, "param", {})
    param.setdefault("filename", "cache.json")

    cache = Cache(**param)
    assert cache.filename == cache_dir / param["filename"]
    assert cache._cache_orig == {}
    assert cache._cache == {}
    assert cache._cache_orig is not cache._cache
    assert not cache._loaded
    assert not cache._dirty
    assert not cache._timer
    assert atexit == [(cache._save, (), {})]

    yield cache

    assert cache._cache_orig is not cache._cache
    assert not cache._timer or not cache._timer.is_alive()


@pytest.mark.usefixtures("no_io")
@pytest.mark.parametrize(
    "filename",
    [
        pytest.param("foo", id="str"),
        pytest.param(Path("foo"), id="Path"),
    ],
)
def test_pathlib_and_str(cache_dir: Path, filename: str | Path):
    cache = Cache(filename)
    assert cache.filename == cache_dir / Path(filename)


@pytest.mark.usefixtures("no_io")
class TestGetterSetter:
    def test_get(self, cache: Cache, mock_schedule_save: Mock):
        assert not mock_schedule_save.called
        assert cache.get("missing-value") is None
        assert cache.get("missing-value", default="default") == "default"
        assert not cache._dirty
        assert mock_schedule_save.call_count == 2

    def test_set(self, cache: Cache):
        assert cache.get("value") is None
        cache.set("value", 1)
        assert cache.get("value") == 1
        assert list(cache._cache.keys()) == ["value"]
        assert cache._dirty

    def test_get_all(self, cache: Cache, mock_schedule_save: Mock):
        assert not mock_schedule_save.called
        cache.set("test1", 1)
        cache.set("test2", 2)
        assert cache.get_all() == {"test1": 1, "test2": 2}
        assert mock_schedule_save.call_count == 3

    def test_get_all_prune(self, cache: Cache):
        cache.set("test1", 1)
        cache.set("test2", 2, -1)
        assert cache.get_all() == {"test1": 1}


@pytest.mark.usefixtures("no_io")
def test_atomicity(monkeypatch: pytest.MonkeyPatch, cache: Cache):
    hs_contains = Handshake()  # block inner Cache.get()
    hs_items = Handshake()  # block inner Cache.get_all()
    hs_setitem = Handshake()  # block inner Cache.set()
    hs_getter = Handshake()
    hs_setter = Handshake()
    hs_getall = Handshake()
    result = None

    class BlockingDict(dict):
        def __contains__(self, key):
            with hs_contains():
                return super().__contains__(key)

        def items(self):
            with hs_items():
                return super().items()

        def __setitem__(self, *args, **kwargs):
            with hs_setitem():
                return super().__setitem__(*args, **kwargs)

    monkeypatch.setattr(cache, "_prune", Mock())
    monkeypatch.setattr(cache, "_cache", BlockingDict(foo={"value": "foo"}))

    def getter_thread():
        nonlocal result
        with hs_getter():
            result = cache.get("foo", "default")

    def setter_thread():
        with hs_setter():
            cache.set("foo", "bar")

    def getall_thread():
        nonlocal result
        with hs_getall():
            result = cache.get_all()

    t_getter = Thread(daemon=True, target=getter_thread)
    t_setter = Thread(daemon=True, target=setter_thread)
    t_getall = Thread(daemon=True, target=getall_thread)

    t_getter.start()
    assert hs_getter.wait_ready(timeout=1), "Getter thread is ready"
    assert not hs_contains.wait_ready(timeout=0), "Getter thread had not yet reached inner get()"

    t_setter.start()
    assert hs_setter.wait_ready(timeout=1), "Setter thread is ready"
    assert not hs_setitem.wait_ready(timeout=0), "Setter thread has not yet reached inner set()"

    # 1. `result = cache.get("foo")`
    hs_getter.go()  # Let getter thread make get() call
    assert hs_contains.wait_ready(timeout=1), "Getter thread has reached inner get()"

    # 2. `cache.set("foo", "bar")` while get() hasn't finished yet
    hs_setter.go()  # Let setter thread make set() call
    assert not hs_setitem.wait_ready(timeout=0.05), "Setter thread won't reach inner set()"

    assert result is None
    assert hs_contains.step(timeout=1), "Let __contains__() return"
    assert hs_getter.wait_done(timeout=1), "Wait for getter thread to finish"
    assert result == "foo"

    assert hs_setitem.wait_ready(timeout=1), "Setter thread has reached inner set() after get() completed"

    t_getall.start()
    assert hs_getall.wait_ready(timeout=1), "Getall thread is ready"
    assert not hs_items.wait_ready(timeout=0), "Getall thread had not yet reached inner get_all()"

    # 3. `result = cache.get_all()` while set() hasn't finished yet
    hs_getall.go()  # Let getall thread make get_all() call
    assert not hs_items.wait_ready(timeout=0.05), "Getall thread won't reach inner get_all()"

    assert hs_setitem.step(timeout=1), "Let __setitem__() return"

    assert hs_items.wait_ready(timeout=1), "Getall thread has reached inner get_all()"
    assert hs_items.step(timeout=1), "Let items() return"
    assert hs_getall.wait_done(timeout=1), "Wait for getall thread to finish"
    assert result == {"foo": "bar"}


@pytest.mark.usefixtures("no_io")
class TestPrefix:
    @pytest.mark.parametrize("cache", [{"key_prefix": "test"}], indirect=["cache"])
    def test_key_prefix(self, cache: Cache):
        cache.set("key", 1)
        assert cache.get("key") == 1
        assert "test:key" in cache._cache
        assert cache._cache["test:key"]["value"] == 1

    def test_get_all_prefix(self, cache: Cache):
        cache.set("test1", 1)
        cache.set("test2", 2)
        cache.key_prefix = "test"
        cache.set("test3", 3)
        cache.set("test4", 4)
        assert cache.get_all() == {"test3": 3, "test4": 4}


@pytest.mark.usefixtures("no_io")
class TestExpiration:
    @pytest.mark.parametrize(
        ("expires", "expected"),
        [
            pytest.param(-20, None, id="past"),
            pytest.param(20, "value", id="future"),
        ],
    )
    def test_expires(self, cache: Cache, expires: float, expected):
        with freezegun.freeze_time("2000-01-01T00:00:00Z"):
            cache.set("key", "value", expires=expires)
            assert cache.get("key") == expected

    @pytest.mark.parametrize(
        ("delta", "expected"),
        [
            pytest.param(timedelta(seconds=-20), None, id="past"),
            pytest.param(timedelta(seconds=20), "value", id="future"),
        ],
    )
    def test_expires_at(self, cache: Cache, delta: timedelta, expected):
        with freezegun.freeze_time("2000-01-01T00:00:00Z"):
            cache.set("key", "value", expires_at=datetime.now(tz=timezone.utc) + delta)
            assert cache.get("key") == expected

    def test_expires_at_overflowerror(self, cache: Cache):
        expires_at = Mock(timestamp=Mock(side_effect=OverflowError))
        cache.set("key", "value", expires_at=expires_at)
        assert cache.get("key") is None

    def test_expiration(self, cache: Cache):
        with freezegun.freeze_time("2000-01-01T00:00:00Z") as frozen_time:
            cache.set("key", "value", expires=20)
            assert cache.get("key") == "value"
            frozen_time.tick(timedelta(seconds=20))
            assert cache.get("key") is None


class TestScheduleSave:
    def test_not_dirty(self, caplog: pytest.LogCaptureFixture, cache: Cache, mock_save: Mock):
        assert not cache._dirty
        cache._schedule_save()
        assert not cache._timer
        assert mock_save.call_count == 0
        assert caplog.records == []

    def test_cancel_timer(self, caplog: pytest.LogCaptureFixture, cache: Cache, mock_save: Mock):
        assert not cache._dirty
        cache._timer = Timer(WRITE_DEBOUNCE_TIME, mock_save)
        cache._timer.daemon = True
        cache._timer.start()
        assert cache._timer.is_alive()
        cache._schedule_save()
        cache._timer.join(timeout=1)
        assert not cache._timer.is_alive()
        assert mock_save.call_count == 0
        assert caplog.records == []

    def test_schedule_save(
        self,
        monkeypatch: pytest.MonkeyPatch,
        caplog: pytest.LogCaptureFixture,
        cache: Cache,
        mock_save: Mock,
    ):
        cache._cache = {"key": {"value": "foo"}}
        assert cache._dirty
        assert not cache._timer

        cache._schedule_save()
        assert mock_save.call_count == 0
        assert cache._timer
        assert cache._timer.is_alive()

        # Let the timer "expire" by monkeypatching the is_set() method and then setting the inner threading.Event flag.
        # This depends on the implementation details of threading.Timer.
        monkeypatch.setattr(cache._timer.finished, "is_set", lambda: False)
        cache._timer.finished.set()  # same as cancel(), but is_set() will return False, so the callback will be executed
        cache._timer.join(timeout=1)
        assert not cache._timer.is_alive()
        assert mock_save.call_count == 1

        assert [(record.name, record.levelname, record.message) for record in caplog.records] == [
            ("streamlink.cache", "trace", "Scheduling write to cache file: 3.0s"),
        ]


class TestIO:
    @pytest.fixture()
    def mock_cache_file(self, request: pytest.FixtureRequest, monkeypatch: pytest.MonkeyPatch):
        content: bytes | Exception = getattr(request, "param", None) or b"{}"
        if not isinstance(content, bytes):
            mock = Mock(side_effect=content)
        else:
            stream = TextIOWrapper(BytesIO(content), encoding="utf-8")
            mock = Mock(return_value=stream)
        monkeypatch.setattr("pathlib.Path.open", mock)

        return mock

    @pytest.mark.parametrize(
        ("mock_cache_file", "mock_json_load", "log"),
        [
            pytest.param(
                FileNotFoundError("File not found"),
                None,
                [],
                id="file-not-found-error",
            ),
            pytest.param(
                PermissionError("Permission denied"),
                None,
                [
                    (
                        "streamlink.cache",
                        "warning",
                        "Failed loading cache file, continuing without cache: Permission denied",
                    ),
                ],
                id="os-error",
            ),
            pytest.param(
                None,
                JSONDecodeError("cache-file.json", "", 0),
                [
                    (
                        "streamlink.cache",
                        "warning",
                        "Failed loading cache file, continuing without cache: cache-file.json: line 1 column 1 (char 0)",
                    ),
                ],
                id="json-decode-error",
            ),
        ],
        indirect=["mock_cache_file", "mock_json_load"],
    )
    def test_load(self, caplog: pytest.LogCaptureFixture, cache: Cache, mock_cache_file: Mock, mock_json_load: Mock, log: list):
        cache._load()
        assert cache._loaded, "Failed load attempts also set the loaded state to True"
        assert not cache._dirty
        assert cache._cache == {}
        assert [(record.name, record.levelname, record.message) for record in caplog.records] == [
            ("streamlink.cache", "trace", f"Loading cache file: {cache.filename}"),
            *log,
        ]
        assert mock_cache_file.call_count == 1

        cache._load()
        assert mock_cache_file.call_count == 1

    @pytest.mark.usefixtures("mock_schedule_save", "mock_save")
    @pytest.mark.parametrize(
        ("mock_cache_file", "expected", "dirty"),
        [
            pytest.param(
                b"""{"foo": {"value": "\\ud83d\\udc3b", "expires": 946684801}}""",
                "🐻",
                False,
                id="load-no-expiration",
            ),
            pytest.param(
                b"""{"foo": {"value": "\\ud83d\\udc3b", "expires": 946684800}}""",
                None,
                True,
                id="load-with-expiration",
            ),
        ],
        indirect=["mock_cache_file"],
    )
    def test_load_data(self, mock_cache_file: Mock, expected: str | None, dirty: bool):
        with freezegun.freeze_time("2000-01-01T00:00:00Z"):
            cache = Cache("mocked-io")
            assert not cache._loaded
            assert not cache._dirty
            assert cache.get("foo") == expected
            assert cache._loaded
            assert cache._dirty is dirty

    @pytest.mark.parametrize(
        ("patch", "side_effect"),
        [
            pytest.param(
                "pathlib.Path.mkdir",
                PermissionError("Permission denied"),
                id="pathlib-path-mkdir",
            ),
            pytest.param(
                "tempfile.NamedTemporaryFile",
                PermissionError("Permission denied"),
                id="tempfile-namedtemporaryfile",
            ),
            pytest.param(
                "json.dump",
                ValueError(),
                id="json-dump",
            ),
            pytest.param(
                "shutil.move",
                PermissionError("Permission denied"),
                id="shutil-move",
            ),
        ],
    )
    def test_save_fail(
        self,
        monkeypatch: pytest.MonkeyPatch,
        caplog: pytest.LogCaptureFixture,
        cache: Cache,
        patch: str,
        side_effect: Exception,
    ):
        mock = Mock(side_effect=side_effect)
        monkeypatch.setattr(patch, mock)

        cache._cache_orig = {"foo": {"value": 1}}
        cache._cache = {"foo": {"value": 2}}
        assert cache._dirty

        cache._save()
        assert mock.call_count == 1
        assert cache._dirty
        assert cache._cache_orig == {"foo": {"value": 1}}
        assert cache._cache == {"foo": {"value": 2}}
        assert not cache.filename.exists()
        assert [(record.name, record.levelname, record.message) for record in caplog.records] == [
            ("streamlink.cache", "trace", f"Writing to cache file: {cache.filename}"),
            ("streamlink.cache", "error", f"Error while writing to cache file: {side_effect}"),
        ]

    @pytest.mark.parametrize("cache", [pytest.param({"filename": Path("foo") / "bar" / "cache.json"})], indirect=True)
    def test_save_success(self, caplog: pytest.LogCaptureFixture, cache: Cache):
        cache._loaded = True
        cache._cache_orig = {"foo": {"value": "bear", "expires": 946684801}}
        cache._cache = {"foo": {"value": "🐻", "expires": 946684801}}
        assert cache._dirty

        cache._save()
        assert not cache._dirty
        assert cache._cache_orig == {"foo": {"value": "🐻", "expires": 946684801}}
        assert cache._cache == cache._cache_orig
        assert cache._cache["foo"] is not cache._cache_orig["foo"]
        assert cache.filename.exists()
        assert (
            cache.filename.read_text(encoding="utf-8")
            == """{\n  "foo": {\n    "value": "\\ud83d\\udc3b",\n    "expires": 946684801\n  }\n}"""
        )
        assert [(record.name, record.levelname, record.message) for record in caplog.records] == [
            ("streamlink.cache", "trace", f"Writing to cache file: {cache.filename}"),
        ]
        with freezegun.freeze_time("2000-01-01T00:00:00Z"):
            assert cache.get("foo") == "🐻"

    def test_save_not_dirty(self, caplog: pytest.LogCaptureFixture, cache: Cache):
        assert not cache._dirty
        cache._save()
        assert not cache.filename.exists()
        assert caplog.records == []

    def test_save_cancel_timer(self, cache: Cache):
        callback = Mock()
        cache._timer = Timer(10, callback)
        cache._timer.daemon = True
        cache._timer.start()
        assert cache._timer.is_alive()
        cache._save()
        cache._timer.join(timeout=1)
        assert not cache._timer.is_alive()
        assert callback.call_count == 0
        assert not cache.filename.exists()

    @pytest.mark.parametrize(
        "mock_cache_file",
        [
            pytest.param(
                b"""{"foo": {"value": "foo", "expires": 946684801}}""",
                id="disabled",
            ),
        ],
        indirect=True,
    )
    def test_disabled(self, caplog: pytest.LogCaptureFixture, mock_cache_file: Mock):
        with freezegun.freeze_time("2000-01-01T00:00:00Z"):
            cache = Cache("mocked-io", disabled=True)
            assert cache._disabled
            assert cache._loaded
            assert cache._cache == {}
            assert cache._cache_orig == {}
            assert mock_cache_file.call_args_list == []

            assert cache.get("foo") is None
            assert not cache._dirty

            cache.set("foo", "bar")
            assert cache._dirty
            assert cache._timer is None

            assert cache.get("foo") == "bar"

            cache._save()
            assert not cache.filename.exists()

        assert [(record.name, record.levelname, record.message) for record in caplog.records] == []



import sys
from pathlib import Path

import pytest

from streamlink.utils.module import load_module


# used in the import test to verify that this module was imported
__test_marker__ = "test_marker"

_here = Path(__file__).parent


@pytest.mark.parametrize(
    ("name", "path", "expected"),
    [
        pytest.param(
            "some_module",
            _here / "does_not_exist",
            ImportError(
                f"Not a package path: {_here / 'does_not_exist'}",
                path=str(_here / "does_not_exist"),
            ),
            id="no-package",
        ),
        pytest.param(
            "does_not_exist",
            _here,
            ImportError(
                "No module named 'does_not_exist'",
                name="does_not_exist",
                path=str(_here),
            ),
            id="no-module",
        ),
    ],
)
def test_load_module_importerror(name: str, path: Path, expected: ImportError):
    with pytest.raises(ImportError) as cm:
        load_module(name, path)
    assert cm.value.msg == expected.msg
    assert cm.value.name == expected.name
    assert cm.value.path == expected.path


def test_load_module():
    name = __name__.split(".")[-1]

    try:
        assert __name__ in sys.modules
        assert name not in sys.modules

        prev = sys.modules[__name__]
        mod1 = load_module(name, Path(__file__).parent)
        assert name in sys.modules
        assert mod1 is sys.modules[name]
        assert prev is sys.modules[__name__]
        assert mod1 is not sys.modules[__name__]

        assert "__test_marker__" in mod1.__dict__
        assert mod1.__test_marker__ == sys.modules[__name__].__test_marker__

        mod2 = load_module(name, Path(__file__).parent)
        assert mod2 is mod1

        mod3 = load_module(name, Path(__file__).parent, override=True)
        assert mod1 is not sys.modules[name]
        assert mod3 is sys.modules[name]
        assert prev is sys.modules[__name__]
        assert mod3 is not sys.modules[__name__]

        assert "__test_marker__" in mod3.__dict__
        assert mod3.__test_marker__ == sys.modules[__name__].__test_marker__

    finally:
        sys.modules.pop(name, None)



from __future__ import annotations

from collections.abc import Mapping
from unittest.mock import Mock

import pytest

from streamlink.stream.stream import Stream
from streamlink_cli.main import resolve_stream_name


@pytest.fixture(scope="module")
def streams():
    a = Stream(Mock())
    b = Stream(Mock())
    c = Stream(Mock())
    d = Stream(Mock())
    e = Stream(Mock())

    return {
        "160p": a,
        "360p": b,
        "480p": c,
        "720p": d,
        "1080p": e,
        "worst": b,
        "best": d,
        "worst-unfiltered": a,
        "best-unfiltered": e,
    }


@pytest.mark.parametrize(
    ("stream_name", "expected"),
    [
        pytest.param("unknown", "unknown"),
        pytest.param("160p", "160p"),
        pytest.param("360p", "360p"),
        pytest.param("480p", "480p"),
        pytest.param("720p", "720p"),
        pytest.param("1080p", "1080p"),
        pytest.param("worst", "360p"),
        pytest.param("best", "720p"),
        pytest.param("worst-unfiltered", "160p"),
        pytest.param("best-unfiltered", "1080p"),
    ],
)
def test_resolve_stream_name(streams: Mapping[str, Stream], stream_name: str, expected: str):
    assert resolve_stream_name(streams, stream_name) == expected



from streamlink.plugins.bbciplayer import BBCiPlayer
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlBBCiPlayer(PluginCanHandleUrl):
    __plugin__ = BBCiPlayer

    should_match_groups = [
        (
            ("live", "http://www.bbc.co.uk/iplayer/live/bbcone"),
            {"channel_name": "bbcone"},
        ),
        (
            ("episode", "http://www.bbc.co.uk/iplayer/episode/b00ymh67/madagascar-1-island-of-marvels"),
            {"episode_id": "b00ymh67"},
        ),
    ]

    should_not_match = [
        "http://www.bbc.co.uk/iplayer/",
    ]


class TestPluginBBCiPlayer:
    def test_vpid_hash(self):
        assert BBCiPlayer._hash_vpid("1234567890") == "71c345435589c6ddeea70d6f252e2a52281ecbf3"



from streamlink.plugins.tvibo import Tvibo
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlTvibo(PluginCanHandleUrl):
    __plugin__ = Tvibo

    should_match = [
        "http://player.tvibo.com/aztv/5929820",
        "http://player.tvibo.com/aztv/6858270/",
        "http://player.tvibo.com/aztv/3977238/",
    ]



from unittest.mock import Mock, call

import pytest

from streamlink.plugins.http import HTTPStreamPlugin
from streamlink.session import Streamlink
from streamlink.stream.http import HTTPStream
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlHTTPStreamPlugin(PluginCanHandleUrl):
    __plugin__ = HTTPStreamPlugin

    should_match_groups = [
        # explicit HTTPStream URLs
        ("httpstream://example.com/foo", {"url": "example.com/foo"}),
        ("httpstream://http://example.com/foo", {"url": "http://example.com/foo"}),
        ("httpstream://https://example.com/foo", {"url": "https://example.com/foo"}),
        # optional parameters
        ("httpstream://example.com/foo abc=def", {"url": "example.com/foo", "params": "abc=def"}),
        ("httpstream://http://example.com/foo abc=def", {"url": "http://example.com/foo", "params": "abc=def"}),
        ("httpstream://https://example.com/foo abc=def", {"url": "https://example.com/foo", "params": "abc=def"}),
    ]

    should_not_match = [
        # missing parameters
        "httpstream://example.com/foo ",
    ]


@pytest.mark.parametrize(
    ("url", "expected"),
    [
        ("httpstream://example.com/foo", "https://example.com/foo"),
        ("httpstream://http://example.com/foo", "http://example.com/foo"),
        ("httpstream://https://example.com/foo", "https://example.com/foo"),
    ],
)
def test_get_streams(
    monkeypatch: pytest.MonkeyPatch,
    session: Streamlink,
    url: str,
    expected: str,
):
    mock_httpstream_init = Mock(return_value=None)
    monkeypatch.setattr("streamlink.stream.http.HTTPStream.__init__", mock_httpstream_init)

    plugin = HTTPStreamPlugin(session, url)
    result = plugin.streams()
    result.pop("worst", None)
    result.pop("best", None)

    assert list(result.keys()) == ["live"]
    assert isinstance(result["live"], HTTPStream)
    assert mock_httpstream_init.call_args_list == [call(session, expected)]


def test_parameters(monkeypatch: pytest.MonkeyPatch, session: Streamlink):
    mock_httpstream_init = Mock(return_value=None)
    monkeypatch.setattr("streamlink.stream.http.HTTPStream.__init__", mock_httpstream_init)

    plugin = HTTPStreamPlugin(
        session,
        (
            "httpstream://example.com/foo"
            + " auth=('foo', 'bar')"
            + " verify=False"
            + " referer=https://example2.com/bar"
            + " params={'key': 'a value'}"
        ),
    )
    plugin.streams()

    assert mock_httpstream_init.call_args_list == [
        call(
            session,
            "https://example.com/foo",
            auth=("foo", "bar"),
            verify=False,
            referer="https://example2.com/bar",
            params={"key": "a value"},
        ),
    ]



from streamlink.plugins.albavision import Albavision
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlAlbavision(PluginCanHandleUrl):
    __plugin__ = Albavision

    should_match_groups = [
        (("antena7", "https://antena7.com.do/envivo-canal-7/"), {}),
        (("antena7", "https://www.antena7.com.do/envivo-canal7"), {}),
        (("antena7", "https://www.antena7.com.do/envivo-canal7/"), {}),
        (("antena7", "https://www.antena7.com.do/envivo-canal7#"), {}),
        (("antena7", "https://www.antena7.com.do/envivo-canal-7#"), {}),
        (("antena7", "https://www.antena7.com.do/en-vivo-canal-99/"), {}),
        (("antena7", "https://www.antena7.com.do/en-vivo-canal-99#"), {}),
        (("antena7", "https://www.antena7.com.do/envivo-canal-7/"), {}),
        (("antena7", "https://www.antena7.com.do/envivo-canal-21/"), {}),
        (("atv", "https://www.atv.pe/envivo-atv"), {}),
        (("atv", "https://www.atv.pe/envivo-atvmas"), {}),
        (("c9n", "https://www.c9n.com.py/envivo/"), {}),
        (("canal10", "https://www.canal10.com.ni/envivo/"), {}),
        (("canal12", "https://www.canal12.com.sv/envivo/"), {}),
        (("chapintv", "https://www.chapintv.com/envivo-canal-3/"), {}),
        (("chapintv", "https://www.chapintv.com/envivo-canal-7/"), {}),
        (("chapintv", "https://www.chapintv.com/envivo-canal-23/"), {}),
        (("elnueve", "https://www.elnueve.com.ar/en-vivo/"), {}),
        (("redbolivision", "https://www.redbolivision.tv.bo/envivo-canal-5/"), {}),
        (("redbolivision", "https://www.redbolivision.tv.bo/upptv/"), {}),
        (("repretel", "https://www.repretel.com/envivo-canal2/"), {}),
        (("repretel", "https://www.repretel.com/envivo-canal4/"), {}),
        (("repretel", "https://www.repretel.com/envivo-canal6/"), {}),
        (("repretel", "https://www.repretel.com/en-vivo-canal-11/"), {}),
        (("rts", "https://www.rts.com.ec/envivo/"), {}),
        (("snt", "https://www.snt.com.py/envivo/"), {}),
        (("tvc", "https://www.tvc.com.ec/envivo/"), {}),
        (("vtv", "https://www.vtv.com.hn/envivo/"), {}),
    ]

    should_not_match = [
        "https://fake.antena7.com.do/envivo-canal-7/",
        "https://www.antena7.com.do/envivo-canal123",
        "https://www.antena7.com.do/envivo-canal123/",
        "https://www.antena7.com.do/envivo-canal-123",
        "https://www.antena7.com.do/envivo-canal-123/",
        "https://www.antena7.com.do/envivo-canal-123#",
        "https://www.antena7.com.do/envivo-canalabc",
        "https://www.antena7.com.do/envivo-canal-abc",
        "https://www.antena7.com.do/envivo-canal-7/extra",
        "https://www.antena7.com.do/envivo-canal-7#extra",
        "https://www.antena7.com.do/something",
    ]


class TestPluginAlbavision:
    def test_transform(self):
        token = Albavision.transform_token("6b425761cc8a86569b1a05a9bf1870c95fca717dOK", 436171)
        assert token == "6b425761cc8a86569b1a05a9bf1870c95fca717d"



from __future__ import annotations

import errno
from collections import deque
from collections.abc import Callable
from pathlib import Path
from threading import Thread
from unittest.mock import Mock, patch

import pytest

from streamlink.stream.stream import StreamIO
from streamlink_cli.console.progress import Progress
from streamlink_cli.output import FileOutput, HTTPOutput, PlayerOutput
from streamlink_cli.streamrunner import PlayerPollThread, StreamRunner, log as streamrunnerlogger
from tests.testutils.handshake import Handshake


TIMEOUT_AWAIT_HANDSHAKE = 1
TIMEOUT_AWAIT_THREADJOIN = 1


class _TestableWithHandshake:
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.handshake = Handshake()


class EventedPlayerPollThread(_TestableWithHandshake, PlayerPollThread):
    POLLING_INTERVAL = 0

    def poll(self):
        with self.handshake():
            return super().poll()

    def close(self):
        super().close()
        # Let thread terminate on close()
        self.handshake.go()


class FakeStream(_TestableWithHandshake, StreamIO):
    """Fake stream implementation, for feeding sample data to the stream runner and simulating read pauses and read errors"""

    def __init__(self) -> None:
        super().__init__()
        self.data: deque[bytes | Callable] = deque()

    # noinspection PyUnusedLocal
    def read(self, *args):
        with self.handshake():
            if not self.data:
                return b""
            data = self.data.popleft()
            return data() if callable(data) else data


class FakeOutput(_TestableWithHandshake):
    """Common output/http-server/progress interface, for caching all write() calls and simulating write errors"""

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.data: list[bytes] = []

    def write(self, data):
        with self.handshake():
            return self._write(data)

    def _write(self, data):
        self.data.append(data)


class FakePlayerOutput(FakeOutput, PlayerOutput):
    def open(self):
        with patch("streamlink_cli.output.player.which", side_effect=lambda path: path):
            return super().open()


class FakeFileOutput(FakeOutput, FileOutput):
    pass


class FakeHTTPOutput(FakeOutput, HTTPOutput):
    pass


class FakeProgress(FakeOutput, Progress):
    # we're not interested in any application logic of the Progress class
    update = print_end = lambda *_, **__: None


class FakeStreamRunner(StreamRunner):
    # override and remove optional typing annotations
    playerpoller: EventedPlayerPollThread
    progress: FakeProgress


@pytest.fixture(autouse=True)
def _logging(caplog: pytest.LogCaptureFixture):
    assert streamrunnerlogger.name == "streamlink.cli"
    caplog.set_level(1, "streamlink")


@pytest.fixture()
def stream():
    stream = FakeStream()
    yield stream
    assert stream.closed


# "stream_runner" fixture dependency declared in downstream scopes
@pytest.fixture()
def runnerthread(request: pytest.FixtureRequest, stream_runner: StreamRunner):
    class RunnerThread(Thread):
        exception = None

        def run(self):
            try:
                super().run()
            except BaseException as err:
                self.exception = err

    thread = RunnerThread(
        daemon=True,
        name="Runner thread",
        target=stream_runner.run,
        args=(b"prebuffer",),
    )
    yield thread

    assert_thread_termination(thread, "Runner thread has terminated")

    exception = getattr(request, "param", {}).get("exception", None)
    assert isinstance(thread.exception, type(exception))
    assert str(thread.exception) == str(exception)


async def assert_handshake_steps(*items: _TestableWithHandshake) -> None:
    for item in items:
        assert await item.handshake.asyncstep(TIMEOUT_AWAIT_HANDSHAKE) is True


def assert_thread_termination(thread: Thread, assertion: str):
    thread.join(TIMEOUT_AWAIT_THREADJOIN)
    assert not thread.is_alive(), assertion


class TestPlayerOutput:
    @pytest.fixture()
    def player_process(self):
        player_process = Mock()
        player_process.poll = Mock(return_value=None)

        return player_process

    @pytest.fixture()
    def output(self, monkeypatch: pytest.MonkeyPatch, player_process: Mock):
        mock_popen = Mock(return_value=player_process)
        monkeypatch.setattr("subprocess.Popen", mock_popen)
        monkeypatch.setattr("streamlink_cli.output.player.sleep", Mock())

        output = FakePlayerOutput(Path("mocked"))
        output.open()
        try:
            yield output
        finally:
            output.close()

    @pytest.fixture()
    def stream_runner(self, monkeypatch: pytest.MonkeyPatch, stream: FakeStream, output: FakePlayerOutput):
        monkeypatch.setattr("streamlink_cli.streamrunner.PlayerPollThread", EventedPlayerPollThread)
        stream_runner = StreamRunner(stream, output)
        assert isinstance(stream_runner.playerpoller, EventedPlayerPollThread)
        assert not stream_runner.playerpoller.is_alive()
        assert not isinstance(stream_runner.output, HTTPOutput)
        assert not stream_runner.progress
        yield stream_runner
        assert not stream_runner.playerpoller.is_alive()

    @pytest.mark.trio()
    async def test_read_write(
        self,
        caplog: pytest.LogCaptureFixture,
        runnerthread: Thread,
        stream_runner: FakeStreamRunner,
        stream: FakeStream,
        output: FakePlayerOutput,
    ):
        stream.data.extend((b"foo", b"bar"))

        runnerthread.start()
        assert output.data == []

        # write prebuffer
        await assert_handshake_steps(output)
        assert output.data == [b"prebuffer"]

        # read and write next chunk
        await assert_handshake_steps(stream, output)
        assert output.data == [b"prebuffer", b"foo"]

        # poll player process
        await assert_handshake_steps(stream_runner.playerpoller)
        assert stream_runner.playerpoller.is_alive()

        # read and write next chunk
        await assert_handshake_steps(stream, output)
        assert output.data == [b"prebuffer", b"foo", b"bar"]

        assert not stream.closed, "Stream is not closed"

        # read stream end
        await assert_handshake_steps(stream)
        assert output.data == [b"prebuffer", b"foo", b"bar"]

        # wait for runner thread to terminate first before asserting log records
        assert_thread_termination(runnerthread, "Runner thread has terminated")
        assert [(record.module, record.levelname, record.message) for record in caplog.records] == [
            ("streamrunner", "info", "Stream ended"),
        ]

    @pytest.mark.trio()
    async def test_paused(
        self,
        caplog: pytest.LogCaptureFixture,
        runnerthread: Thread,
        stream_runner: FakeStreamRunner,
        stream: FakeStream,
        output: FakePlayerOutput,
    ):
        delayed = Handshake()

        def item():
            with delayed():
                return b"delayed"

        stream.data.append(item)

        runnerthread.start()
        assert output.data == []

        # write prebuffer
        await assert_handshake_steps(output)
        assert output.data == [b"prebuffer"]
        assert not delayed.wait_ready(0), "Delayed chunk has not been read yet"

        # attempt reading delayed chunk
        stream.handshake.go()
        assert delayed.wait_ready(TIMEOUT_AWAIT_HANDSHAKE), "read() call of delayed chunk is paused"
        assert output.data == [b"prebuffer"]

        assert not stream.closed, "Stream is not closed"

        # poll player process
        await assert_handshake_steps(stream_runner.playerpoller)
        assert stream_runner.playerpoller.is_alive()

        # unpause delayed chunk
        delayed.go()
        assert stream.handshake.wait_done(TIMEOUT_AWAIT_HANDSHAKE), "Delayed chunk has successfully been read"
        await assert_handshake_steps(output)
        assert output.data == [b"prebuffer", b"delayed"]

        assert not stream.closed, "Stream is not closed"

        # read stream end
        await assert_handshake_steps(stream)
        assert output.data == [b"prebuffer", b"delayed"]

        # wait for runner thread to terminate first before asserting log records
        assert_thread_termination(runnerthread, "Runner thread has terminated")
        assert [(record.module, record.levelname, record.message) for record in caplog.records] == [
            ("streamrunner", "info", "Stream ended"),
        ]

    @pytest.mark.trio()
    @pytest.mark.parametrize(
        ("writeerror", "runnerthread"),
        [
            pytest.param(
                OSError(errno.EPIPE, "Broken pipe"),
                {},
                id="Acceptable error: EPIPE",
            ),
            pytest.param(
                OSError(errno.EINVAL, "Invalid argument"),
                {},
                id="Acceptable error: EINVAL",
            ),
            pytest.param(
                OSError(errno.ECONNRESET, "Connection reset"),
                {},
                id="Acceptable error: ECONNRESET",
            ),
            pytest.param(
                OSError("Unknown error"),
                {"exception": OSError("Error when writing to output: Unknown error, exiting")},
                id="Non-acceptable error",
            ),
        ],
        indirect=["runnerthread"],
    )
    async def test_player_close(
        self,
        caplog: pytest.LogCaptureFixture,
        runnerthread: Thread,
        stream_runner: FakeStreamRunner,
        stream: FakeStream,
        output: FakePlayerOutput,
        player_process: Mock,
        writeerror: Exception,
    ):
        stream.data.extend((b"foo", b"bar"))

        runnerthread.start()
        assert output.data == []

        # write prebuffer
        await assert_handshake_steps(output)
        assert output.data == [b"prebuffer"]

        # poll player process
        await assert_handshake_steps(stream_runner.playerpoller)
        assert stream_runner.playerpoller.is_alive()

        # read and write next chunk
        await assert_handshake_steps(stream, output)
        assert output.data == [b"prebuffer", b"foo"]

        assert not stream.closed, "Stream is not closed yet"

        # close player
        with patch.object(output, "_write", side_effect=writeerror):
            # let player process terminate with code 0 and poll process once
            player_process.poll.return_value = 0
            await assert_handshake_steps(stream_runner.playerpoller)
            assert_thread_termination(stream_runner.playerpoller, "Polling has stopped after player process terminated")

            assert stream.closed, "Stream got closed after the player was closed"

            # read and write next chunk (write will now also raise)
            await assert_handshake_steps(stream, output)
            assert output.data == [b"prebuffer", b"foo"]

        # wait for runner thread to terminate first before asserting log records
        assert_thread_termination(runnerthread, "Runner thread has terminated")
        assert [(record.module, record.levelname, record.message) for record in caplog.records] == [
            ("streamrunner", "info", "Player closed"),
            ("streamrunner", "info", "Stream ended"),
        ]

    @pytest.mark.trio()
    async def test_player_close_paused(
        self,
        caplog: pytest.LogCaptureFixture,
        runnerthread: Thread,
        stream_runner: FakeStreamRunner,
        stream: FakeStream,
        output: FakePlayerOutput,
        player_process: Mock,
    ):
        delayed = Handshake()

        def item():
            with delayed():
                return b""

        stream.data.append(item)

        runnerthread.start()
        assert output.data == []

        # write prebuffer
        await assert_handshake_steps(output)
        assert output.data == [b"prebuffer"]
        assert not delayed.wait_ready(0), "Delayed chunk has not been read yet"

        # poll player process
        await assert_handshake_steps(stream_runner.playerpoller)
        assert stream_runner.playerpoller.is_alive()

        stream.handshake.go()
        assert delayed.wait_ready(TIMEOUT_AWAIT_HANDSHAKE), "read() call of delayed chunk is paused"
        assert output.data == [b"prebuffer"]

        assert not stream.closed, "Stream is not closed yet"

        # let player process terminate with code 0 and poll process once
        player_process.poll.return_value = 0
        await assert_handshake_steps(stream_runner.playerpoller)
        assert_thread_termination(stream_runner.playerpoller, "Polling has stopped after player process terminated")

        assert stream.closed, "Stream got closed after the player was closed, even if the stream was paused"

        # unpause delayed chunk (stream end)
        delayed.go()
        assert stream.handshake.wait_done(TIMEOUT_AWAIT_HANDSHAKE), "Delayed chunk has successfully been read"
        assert output.data == [b"prebuffer"]

        # wait for runner thread to terminate first before asserting log records
        assert_thread_termination(runnerthread, "Runner thread has terminated")
        assert [(record.module, record.levelname, record.message) for record in caplog.records] == [
            ("streamrunner", "info", "Player closed"),
            ("streamrunner", "info", "Stream ended"),
        ]

    @pytest.mark.trio()
    @pytest.mark.parametrize(
        "runnerthread",
        [{"exception": OSError("Error when reading from stream: Read timeout, exiting")}],
        indirect=["runnerthread"],
    )
    async def test_readerror(
        self,
        caplog: pytest.LogCaptureFixture,
        runnerthread: Thread,
        stream_runner: FakeStreamRunner,
        stream: FakeStream,
        output: FakePlayerOutput,
    ):
        # make next read() call raise a read-timeout error
        stream.data.append(Mock(side_effect=OSError("Read timeout")))

        runnerthread.start()
        assert output.data == []

        # write prebuffer
        await assert_handshake_steps(output)
        assert output.data == [b"prebuffer"]

        # poll player process
        await assert_handshake_steps(stream_runner.playerpoller)
        assert stream_runner.playerpoller.is_alive()

        # read stream (will raise a read timeout)
        await assert_handshake_steps(stream)

        # poll player process again
        await assert_handshake_steps(stream_runner.playerpoller)
        assert_thread_termination(stream_runner.playerpoller, "Polling has stopped on read error")

        # wait for runner thread to terminate first before asserting log records
        assert_thread_termination(runnerthread, "Runner thread has terminated")
        assert [(record.module, record.levelname, record.message) for record in caplog.records] == [
            ("streamrunner", "info", "Stream ended"),
        ]


class TestHTTPServer:
    @pytest.fixture()
    def output(self):
        return FakeHTTPOutput()

    @pytest.fixture()
    def stream_runner(self, stream: FakeStream, output: FakeHTTPOutput):
        stream_runner = StreamRunner(stream, output)
        assert not stream_runner.playerpoller
        assert not stream_runner.progress
        assert isinstance(stream_runner.output, HTTPOutput)
        return stream_runner

    @pytest.mark.trio()
    async def test_read_write(
        self,
        caplog: pytest.LogCaptureFixture,
        runnerthread: Thread,
        stream_runner: FakeStreamRunner,
        stream: FakeStream,
        output: FakeHTTPOutput,
    ):
        stream.data.extend((b"foo", b"bar"))

        runnerthread.start()
        assert output.data == []

        # write prebuffer
        await assert_handshake_steps(output)
        assert output.data == [b"prebuffer"]

        # read and write next chunk
        await assert_handshake_steps(stream, output)
        assert output.data == [b"prebuffer", b"foo"]

        # read and write next chunk
        await assert_handshake_steps(stream, output)
        assert output.data == [b"prebuffer", b"foo", b"bar"]

        assert not stream.closed, "Stream is not closed"

        # read stream end
        await assert_handshake_steps(stream)
        assert output.data == [b"prebuffer", b"foo", b"bar"]

        # wait for runner thread to terminate first before asserting log records
        assert_thread_termination(runnerthread, "Runner thread has terminated")
        assert [(record.module, record.levelname, record.message) for record in caplog.records] == [
            ("streamrunner", "info", "Stream ended"),
        ]

    @pytest.mark.parametrize(
        ("writeerror", "logs", "runnerthread"),
        [
            pytest.param(
                OSError(errno.EPIPE, "Broken pipe"),
                True,
                {},
                id="Acceptable error: EPIPE",
            ),
            pytest.param(
                OSError(errno.EINVAL, "Invalid argument"),
                True,
                {},
                id="Acceptable error: EINVAL",
            ),
            pytest.param(
                OSError(errno.ECONNRESET, "Connection reset"),
                True,
                {},
                id="Acceptable error: ECONNRESET",
            ),
            pytest.param(
                OSError("Unknown error"),
                False,
                {"exception": OSError("Error when writing to output: Unknown error, exiting")},
                id="Non-acceptable error",
            ),
        ],
        indirect=["runnerthread"],
    )
    def test_writeerror(
        self,
        caplog: pytest.LogCaptureFixture,
        runnerthread: Thread,
        stream_runner: FakeStreamRunner,
        stream: FakeStream,
        output: FakePlayerOutput,
        logs: bool,
        writeerror: Exception,
    ):
        runnerthread.start()

        with patch.object(output, "_write", side_effect=writeerror):
            assert output.handshake.step(TIMEOUT_AWAIT_HANDSHAKE)
            assert output.data == []

        # wait for runner thread to terminate first before asserting log records
        assert_thread_termination(runnerthread, "Runner thread has terminated")
        expectedlogs = (
            ([("streamrunner", "info", "HTTP connection closed")] if logs else [])
            + [("streamrunner", "info", "Stream ended")]
        )  # fmt: skip
        assert [(record.module, record.levelname, record.message) for record in caplog.records] == expectedlogs


class TestProgress:
    @pytest.fixture()
    def output(self):
        return FakeFileOutput(Path("filename"))

    @pytest.fixture()
    def progress(self):
        return FakeProgress(console=Mock(), path=Path("filename"))

    @pytest.fixture()
    def stream_runner(self, stream: FakeStream, output: FakeFileOutput, progress: FakeProgress):
        stream_runner = FakeStreamRunner(stream, output, progress=progress)
        assert not stream_runner.playerpoller
        assert stream_runner.progress is progress
        assert not stream_runner.progress.is_alive()
        yield stream_runner
        assert not stream_runner.progress.is_alive()

    @pytest.mark.trio()
    async def test_read_write(
        self,
        caplog: pytest.LogCaptureFixture,
        runnerthread: Thread,
        stream_runner: FakeStreamRunner,
        stream: FakeStream,
        output: FakeFileOutput,
    ):
        stream.data.extend((b"foo", b"bar"))

        runnerthread.start()
        assert output.data == []

        # write prebuffer
        await assert_handshake_steps(output, stream_runner.progress)
        assert output.data == [b"prebuffer"]
        assert stream_runner.progress.data == [b"prebuffer"]

        # read and write next chunk
        await assert_handshake_steps(stream, output, stream_runner.progress)
        assert output.data == [b"prebuffer", b"foo"]
        assert stream_runner.progress.data == [b"prebuffer", b"foo"]

        # read and write next chunk
        await assert_handshake_steps(stream, output, stream_runner.progress)
        assert output.data == [b"prebuffer", b"foo", b"bar"]
        assert stream_runner.progress.data == [b"prebuffer", b"foo", b"bar"]

        assert not stream.closed, "Stream is not closed"

        # read stream end
        await assert_handshake_steps(stream)
        assert output.data == [b"prebuffer", b"foo", b"bar"]
        assert stream_runner.progress.data == [b"prebuffer", b"foo", b"bar"]

        # wait for runner thread to terminate first before asserting log records
        assert_thread_termination(runnerthread, "Runner thread has terminated")
        assert [(record.module, record.levelname, record.message) for record in caplog.records] == [
            ("streamrunner", "info", "Stream ended"),
        ]
